from __future__ import annotations

import time

import fsspec
import pandas as pd
import pyarrow as pa
import pytest
from aiohttp.client_exceptions import ClientPayloadError
from pyarrow import parquet as pq

import daft
import daft.recordbatch
from daft.exceptions import ConnectTimeoutError, ReadTimeoutError
from daft.filesystem import get_filesystem, get_protocol_from_path
from daft.recordbatch import MicroPartition, RecordBatch


def get_filesystem_from_path(path: str, **kwargs) -> fsspec.AbstractFileSystem:
    protocol = get_protocol_from_path(path)
    fs = get_filesystem(protocol, **kwargs)
    return fs


# TODO(desmond): Expected some of these non-working cases to work after recent fixes, but that
#                doesn't seem to be the case.
# Taken from our spreadsheet of files that Daft should be able to handle
DAFT_CAN_READ_FILES = [
    (
        "parquet-testing/data/alltypes_dictionary.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/alltypes_dictionary.parquet",
    ),
    (
        "parquet-testing/data/alltypes_plain.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/alltypes_plain.parquet",
    ),
    (
        "parquet-testing/data/alltypes_plain.snappy.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/alltypes_plain.snappy.parquet",
    ),
    (
        "parquet-testing/data/alltypes_tiny_pages.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/alltypes_tiny_pages.parquet",
    ),
    (
        "parquet-testing/data/alltypes_tiny_pages_plain.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/alltypes_tiny_pages_plain.parquet",
    ),
    (
        "parquet-testing/data/binary.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/binary.parquet",
    ),
    # Needs Decimals decoding from byte arrays
    (
        "parquet-testing/data/byte_array_decimal.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/byte_array_decimal.parquet",
    ),
    (
        "parquet-testing/data/data_index_bloom_encoding_stats.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/data_index_bloom_encoding_stats.parquet",
    ),
    (
        "parquet-testing/data/datapage_v1-snappy-compressed-checksum.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/datapage_v1-snappy-compressed-checksum.parquet",
    ),
    (
        "parquet-testing/data/datapage_v1-uncompressed-checksum.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/datapage_v1-uncompressed-checksum.parquet",
    ),
    # Thrift Error?
    # (
    #     "parquet-testing/data/dict-page-offset-zero.parquet",
    #     "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/dict-page-offset-zero.parquet",
    # ),
    # Need Fixed Length Binary in Daft or convert to Variable Sized Binary
    # (
    #     "parquet-testing/data/fixed_length_byte_array.parquet",
    #     "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/fixed_length_byte_array.parquet",
    # ),
    (
        "parquet-testing/data/fixed_length_decimal.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/fixed_length_decimal.parquet",
    ),
    (
        "parquet-testing/data/fixed_length_decimal_legacy.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/fixed_length_decimal_legacy.parquet",
    ),
    (
        "parquet-testing/data/int32_decimal.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/int32_decimal.parquet",
    ),
    (
        "parquet-testing/data/int32_with_null_pages.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/int32_with_null_pages.parquet",
    ),
    (
        "parquet-testing/data/int64_decimal.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/int64_decimal.parquet",
    ),
    (
        "parquet-testing/data/list_columns.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/list_columns.parquet",
    ),
    (
        "parquet-testing/data/nan_in_stats.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/nan_in_stats.parquet",
    ),
    # Page Header Wrong Size?
    # (
    #     "parquet-testing/data/nation.dict-malformed.parquet",
    #     "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/nation.dict-malformed.parquet",
    # ),
    (
        "parquet-testing/data/nested_lists.snappy.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/nested_lists.snappy.parquet",
    ),
    # We have problems decoding struct objects in our arrow2 decoder
    # (
    #     "parquet-testing/data/nested_structs.rust.parquet",
    #     "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/nested_structs.rust.parquet",
    # ),
    # We currently don't support Map Dtypes
    # (
    #     "parquet-testing/data/nonnullable.impala.parquet",
    #     "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/nonnullable.impala.parquet",
    # ),
    (
        "parquet-testing/data/null_list.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/null_list.parquet",
    ),
    # We currently don't support Map Dtypes
    # (
    #     "parquet-testing/data/nullable.impala.parquet",
    #     "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/nullable.impala.parquet",
    # ),
    (
        "parquet-testing/data/nulls.snappy.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/nulls.snappy.parquet",
    ),
    # For some reason the program segfaults with this file unless we make the chunk size > 2024
    # (
    #     "parquet-testing/data/overflow_i16_page_cnt.parquet",
    #     "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/overflow_i16_page_cnt.parquet",
    # ),
    (
        "parquet-testing/data/plain-dict-uncompressed-checksum.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/plain-dict-uncompressed-checksum.parquet",
    ),
    # We have problems decoding struct objects in our arrow2 decoder
    # (
    #     "parquet-testing/data/repeated_no_annotation.parquet",
    #     "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/repeated_no_annotation.parquet",
    # ),
    (
        "parquet-testing/data/rle-dict-snappy-checksum.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/rle-dict-snappy-checksum.parquet",
    ),
    # We currently don't support RLE Boolean encodings
    # (
    #     "parquet-testing/data/rle_boolean_encoding.parquet",
    #     "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/rle_boolean_encoding.parquet",
    # ),
    (
        "parquet-testing/data/single_nan.parquet",
        "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/github/apache/parquet-testing/data/single_nan.parquet",
    ),
    # This is currently is in a private s3 bucket
    # (
    #     "daft-tpch/100g_32part",
    #     "s3://eventual-dev-benchmarking-fixtures/uncompressed/tpch-dbgen/100_0/32/parquet/lineitem/108417bd-5bee-43d9-bf9a-d6faec6afb2d-0.parquet",
    # ),
    (
        "parquet-benchmarking/mvp",
        "s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet",
    ),
    (
        "parquet-benchmarking/s3a-mvp",
        "s3a://daft-public-data/test_fixtures/parquet-dev/mvp.parquet",
    ),
    (
        "azure/mvp/az",
        "az://public-anonymous/mvp.parquet",
    ),
    (
        "azure/mvp/abfs",
        "abfs://public-anonymous/mvp.parquet",
    ),
    (
        "azure/mvp/abfss",
        "abfss://public-anonymous/mvp.parquet",
    ),
    (
        "gcs/mvp",
        "gs://daft-public-data-gs/mvp.parquet",
    ),
    (
        "daft/schema_with_metadata",
        "tests/assets/parquet-data/parquet-with-schema-metadata.parquet",
    ),
]


@pytest.fixture(scope="session")
def public_storage_io_config() -> daft.io.IOConfig:
    return daft.io.IOConfig(
        azure=daft.io.AzureConfig(storage_account="dafttestdata", anonymous=True),
        s3=daft.io.S3Config(region_name="us-west-2", anonymous=True),
        gcs=daft.io.GCSConfig(anonymous=True),
    )


@pytest.fixture(scope="session", params=DAFT_CAN_READ_FILES, ids=[name for name, _ in DAFT_CAN_READ_FILES])
def parquet_file(request) -> tuple[str, str]:
    """Returns a tuple of (`name`, `url`) of files that Daft should be able to handle. URLs may be HTTPs or S3."""
    return request.param


@pytest.fixture(params=[(True, True), (True, False), (False, False)], ids=["split", "merge", "ignore"])
def set_split_config(request):
    max_size = 0 if request.param[0] else 384 * 1024 * 1024
    min_size = 0 if request.param[1] else 96 * 1024 * 1024

    with daft.execution_config_ctx(
        scan_tasks_max_size_bytes=max_size,
        scan_tasks_min_size_bytes=min_size,
    ):
        yield


def read_parquet_with_pyarrow(path) -> pa.Table:
    kwargs = {}
    if get_protocol_from_path(path) == "s3" or get_protocol_from_path(path) == "s3a":
        kwargs["anon"] = True
    if get_protocol_from_path(path) in ("az", "abfs", "abfss"):
        kwargs["account_name"] = "dafttestdata"
        kwargs["anon"] = True

    fs = get_filesystem_from_path(path, **kwargs)

    table = None
    retries = 0
    while not table:
        try:
            table = pq.read_table(path, filesystem=fs)
        except ClientPayloadError as e:
            if retries > 5:
                raise e
            time.sleep(0.1 * (2**retries))
            retries += 1
    return table


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_parquet_read_table(parquet_file, public_storage_io_config, multithreaded_io):
    _, url = parquet_file
    daft_native_read = MicroPartition.read_parquet(
        url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io
    )
    pa_read = MicroPartition.from_arrow(read_parquet_with_pyarrow(url))
    assert daft_native_read.schema() == pa_read.schema()
    pd.testing.assert_frame_equal(daft_native_read.to_pandas(), pa_read.to_pandas())


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_parquet_read_table_into_pyarrow(parquet_file, public_storage_io_config, multithreaded_io):
    _, url = parquet_file
    daft_native_read = daft.recordbatch.read_parquet_into_pyarrow(
        url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io
    )
    pa_read = read_parquet_with_pyarrow(url)
    assert daft_native_read.schema == pa_read.schema
    assert pa_read.schema.metadata is None or daft_native_read.schema.metadata == pa_read.schema.metadata
    pd.testing.assert_frame_equal(daft_native_read.to_pandas(), pa_read.to_pandas())


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_parquet_read_table_bulk(parquet_file, public_storage_io_config, multithreaded_io):
    _, url = parquet_file
    daft_native_reads = MicroPartition.read_parquet_bulk(
        [url] * 2, io_config=public_storage_io_config, multithreaded_io=multithreaded_io
    )
    pa_read = MicroPartition.from_arrow(read_parquet_with_pyarrow(url))

    # Legacy Table returns a list[Table]
    if MicroPartition == RecordBatch:
        for daft_native_read in daft_native_reads:
            assert daft_native_read.schema() == pa_read.schema()
            pd.testing.assert_frame_equal(daft_native_read.to_pandas(), pa_read.to_pandas())
    # MicroPartitions returns a MicroPartition
    else:
        assert daft_native_reads.schema() == pa_read.schema()
        pd.testing.assert_frame_equal(
            daft_native_reads.to_pandas(), MicroPartition.concat([pa_read, pa_read]).to_pandas()
        )


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_parquet_into_pyarrow_bulk(parquet_file, public_storage_io_config, multithreaded_io):
    _, url = parquet_file
    daft_native_reads = daft.recordbatch.read_parquet_into_pyarrow_bulk(
        [url] * 2, io_config=public_storage_io_config, multithreaded_io=multithreaded_io
    )
    pa_read = read_parquet_with_pyarrow(url)

    for daft_native_read in daft_native_reads:
        assert daft_native_read.schema == pa_read.schema
        pd.testing.assert_frame_equal(daft_native_read.to_pandas(), pa_read.to_pandas())


@pytest.mark.integration()
def test_parquet_read_df(parquet_file, public_storage_io_config, set_split_config):
    _, url = parquet_file
    daft_native_read = daft.read_parquet(url, io_config=public_storage_io_config)
    pa_read = MicroPartition.from_arrow(read_parquet_with_pyarrow(url))
    assert daft_native_read.schema() == pa_read.schema()
    pd.testing.assert_frame_equal(daft_native_read.to_pandas(), pa_read.to_pandas())


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_row_groups_selection(public_storage_io_config, multithreaded_io):
    url = "s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"
    all_rows = MicroPartition.read_parquet(url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io)
    assert len(all_rows) == 100
    first = MicroPartition.read_parquet(
        url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io, row_groups=[0]
    )
    assert len(first) == 10
    assert all_rows.to_arrow()[:10] == first.to_arrow()

    fifth = MicroPartition.read_parquet(
        url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io, row_groups=[5]
    )
    assert len(fifth) == 10
    assert all_rows.to_arrow()[50:60] == fifth.to_arrow()

    repeated = MicroPartition.read_parquet(
        url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io, row_groups=[1, 1, 1]
    )
    assert len(repeated) == 30
    assert all_rows.to_arrow()[10:20] == repeated.to_arrow()[:10]
    assert all_rows.to_arrow()[10:20] == repeated.to_arrow()[10:20]
    assert all_rows.to_arrow()[10:20] == repeated.to_arrow()[20:]

    out_of_order = MicroPartition.read_parquet(
        url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io, row_groups=[1, 0]
    )
    assert len(out_of_order) == 20
    assert all_rows.to_arrow()[10:20] == out_of_order.to_arrow()[:10]
    assert all_rows.to_arrow()[0:10] == out_of_order.to_arrow()[10:20]


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_row_groups_selection_bulk(public_storage_io_config, multithreaded_io):
    url = ["s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"] * 11
    row_groups = [list(range(10))] + [[i] for i in range(10)]

    if MicroPartition == RecordBatch:
        first, *rest = MicroPartition.read_parquet_bulk(
            url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io, row_groups_per_path=row_groups
        )
        assert len(first) == 100
        assert len(rest) == 10

        for i, t in enumerate(rest):
            assert len(t) == 10
            assert first.to_arrow()[i * 10 : (i + 1) * 10] == t.to_arrow()
    else:
        mp = MicroPartition.read_parquet_bulk(
            url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io, row_groups_per_path=row_groups
        )
        assert len(mp) == 100 + (
            10 * 10
        )  # 100 rows in first table (10 rgs), 10 rows each in subsequent tables (1 rg each)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_row_groups_selection_into_pyarrow_bulk(public_storage_io_config, multithreaded_io):
    url = ["s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"] * 11
    row_groups = [list(range(10))] + [[i] for i in range(10)]
    first, *rest = daft.recordbatch.read_parquet_into_pyarrow_bulk(
        url, io_config=public_storage_io_config, multithreaded_io=multithreaded_io, row_groups_per_path=row_groups
    )
    assert len(first) == 100
    assert len(rest) == 10

    for i, t in enumerate(rest):
        assert len(t) == 10
        assert first[i * 10 : (i + 1) * 10] == t


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_connect_timeout_s3(multithreaded_io):
    url = "s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"
    connect_timeout_config = daft.io.IOConfig(
        s3=daft.io.S3Config(
            # NOTE: no keys or endpoints specified for an AWS public s3 bucket
            region_name="us-west-2",
            anonymous=True,
            connect_timeout_ms=1,
            num_tries=3,
        )
    )

    with pytest.raises((ReadTimeoutError, ConnectTimeoutError), match=f"timed out when trying to connect to {url}"):
        MicroPartition.read_parquet(url, io_config=connect_timeout_config, multithreaded_io=multithreaded_io).to_arrow()


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_read_timeout_s3(multithreaded_io):
    url = "s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"
    read_timeout_config = daft.io.IOConfig(
        s3=daft.io.S3Config(
            # NOTE: no keys or endpoints specified for an AWS public s3 bucket
            region_name="us-west-2",
            anonymous=True,
            read_timeout_ms=1,
            num_tries=3,
        )
    )

    with pytest.raises((ReadTimeoutError, ConnectTimeoutError), match=f"timed out when trying to connect to {url}"):
        MicroPartition.read_parquet(url, io_config=read_timeout_config, multithreaded_io=multithreaded_io).to_arrow()


@pytest.mark.integration()
def test_read_file_level_timeout():
    url = "s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"
    read_timeout_config = daft.io.IOConfig(
        s3=daft.io.S3Config(
            # NOTE: no keys or endpoints specified for an AWS public s3 bucket
            region_name="us-west-2",
            anonymous=True,
            num_tries=3,
        )
    )

    with pytest.raises((ReadTimeoutError), match=f"Parquet reader timed out while trying to read: {url}"):
        daft.recordbatch.read_parquet_into_pyarrow(url, io_config=read_timeout_config, file_timeout_ms=2)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_connect_timeout_gcs(multithreaded_io):
    url = "gs://daft-public-data-gs/mvp.parquet"
    connect_timeout_config = daft.io.IOConfig(
        gcs=daft.io.GCSConfig(
            anonymous=True,
            connect_timeout_ms=1,
            retry_initial_backoff_ms=10,
            num_tries=3,
        )
    )

    with pytest.raises((ReadTimeoutError, ConnectTimeoutError), match=f"timed out when trying to connect to {url}"):
        MicroPartition.read_parquet(url, io_config=connect_timeout_config, multithreaded_io=multithreaded_io).to_arrow()


@pytest.mark.integration()
@pytest.mark.parametrize(
    "multithreaded_io",
    [False, True],
)
def test_read_timeout_gcs(multithreaded_io):
    url = "gs://daft-public-data-gs/mvp.parquet"
    read_timeout_config = daft.io.IOConfig(
        gcs=daft.io.GCSConfig(
            anonymous=True,
            read_timeout_ms=1,
            retry_initial_backoff_ms=10,
            num_tries=3,
        )
    )

    with pytest.raises((ReadTimeoutError, ConnectTimeoutError), match=f"Read timed out when trying to read {url}"):
        MicroPartition.read_parquet(url, io_config=read_timeout_config, multithreaded_io=multithreaded_io).to_arrow()



from __future__ import annotations

import pytest

from daft.expressions import col
from daft.expressions.expressions import lit
from daft.recordbatch.micropartition import MicroPartition


@pytest.mark.parametrize(
    ["predicate", "if_true", "if_false", "expected"],
    [
        # Single row
        ([True], [1], [2], [1]),
        ([False], [1], [2], [2]),
        # Multiple rows
        ([True, False, True], [1, 2, 3], [4, 5, 6], [1, 5, 3]),
        ([False, False, False], [1, 2, 3], [4, 5, 6], [4, 5, 6]),
    ],
)
def test_table_expr_if_else(predicate, if_true, if_false, expected) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"predicate": predicate, "if_true": if_true, "if_false": if_false})
    daft_recordbatch = daft_recordbatch.eval_expression_list(
        [col("predicate").if_else(col("if_true"), col("if_false"))]
    )
    pydict = daft_recordbatch.to_pydict()

    assert pydict["if_true"] == expected


@pytest.mark.skip(reason="missing_key does not get evaluated but it errors on binding")
@pytest.mark.parametrize(
    "if_else_expr",
    [
        lit(True).if_else(col("struct").struct.get("key"), col("struct").struct.get("missing_key")),
        lit(False).if_else(col("struct").struct.get("missing_key"), col("struct").struct.get("key")),
    ],
)
def test_table_expr_if_else_literal_predicate(if_else_expr) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"struct": [{"key": "value"}]})
    daft_recordbatch = daft_recordbatch.eval_expression_list([if_else_expr])
    pydict = daft_recordbatch.to_pydict()

    assert pydict == {if_else_expr.name(): ["value"]}



from __future__ import annotations

import daft


def test_warc():
    path = "tests/assets/example.warc"
    df = daft.read_warc(path)
    assert df.count_rows() == 30
    assert df.filter(df["WARC-Type"] == "response").count_rows() == 11
    # Test that we can correctly extract metadata from the warc_headers json column.
    num_rows = (
        df.with_column("WARC-Identified-Payload-Type", daft.col("warc_headers").json.query('."Content-Type"'))
        .where(
            (daft.col("WARC-Type") == "response")
            & (daft.col("Content-Length") < 10 * 1024 * 1024)
            & (daft.col("WARC-Identified-Payload-Type").is_in(['"application/http; msgtype=response"']))
        )
        .count_rows()
    )
    assert num_rows == 11


def test_warc_gz():
    # Test that we can read a gzipped warc file.
    path = "tests/assets/example.warc.gz"
    df = daft.read_warc(path)
    assert df.count_rows() == 30
    assert df.filter(df["WARC-Type"] == "response").count_rows() == 11
    # Test that we can correctly extract metadata from the warc_headers json column.
    num_rows = (
        df.with_column("WARC-Identified-Payload-Type", daft.col("warc_headers").json.query('."Content-Type"'))
        .where(
            (daft.col("WARC-Type") == "response")
            & (daft.col("Content-Length") < 10 * 1024 * 1024)
            & (daft.col("WARC-Identified-Payload-Type").is_in(['"application/http; msgtype=response"']))
        )
        .count_rows()
    )
    assert num_rows == 11



from __future__ import annotations

from typing import TYPE_CHECKING

from daft.daft import FileFormatConfig, ScanOperatorHandle, StorageConfig
from daft.logical.builder import LogicalPlanBuilder
from daft.logical.schema import Schema

if TYPE_CHECKING:
    from daft.datatype import DataType


def _get_schema_from_dict(fields: dict[str, DataType]) -> Schema:
    if isinstance(fields, dict):
        return Schema._from_field_name_and_types([(fname, dtype) for fname, dtype in fields.items()])
    else:
        raise NotImplementedError(f"Unsupported schema fields: {type(fields)}")


def get_tabular_files_scan(
    path: str | list[str],
    infer_schema: bool,
    schema: dict[str, DataType] | None,
    file_format_config: FileFormatConfig,
    storage_config: StorageConfig,
    file_path_column: str | None = None,
    hive_partitioning: bool = False,
) -> LogicalPlanBuilder:
    """Returns a TabularFilesScan LogicalPlan for a given glob filepath."""
    # Glob the path using the Runner

    if isinstance(path, list):
        paths = path
    elif isinstance(path, str):
        paths = [path]
    else:
        raise NotImplementedError(f"get_tabular_files_scan cannot construct ScanOperatorHandle for input: {path}")

    scan_op = ScanOperatorHandle.glob_scan(
        paths,
        file_format_config,
        storage_config,
        infer_schema=infer_schema,
        schema=_get_schema_from_dict(schema)._schema if schema is not None else None,
        file_path_column=file_path_column,
        hive_partitioning=hive_partitioning,
    )

    builder = LogicalPlanBuilder.from_tabular_scan(
        scan_operator=scan_op,
    )
    return builder



from __future__ import annotations

import daft
from daft.datatype import DataType
from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_list_sort():
    table = MicroPartition.from_pydict(
        {
            "a": [[1, 3], [4, 6, 2], [3], [11, 6], None, [], [1, None, 3], None],
            "b": [True, False, False, True, False, False, False, True],
        }
    ).eval_expression_list([col("a").cast(DataType.list(DataType.int64())), col("b")])

    res = table.eval_expression_list(
        [
            col("a").list.sort().alias("asc"),
            col("a").list.sort(True).alias("desc"),
            col("a").list.sort(col("b")).alias("mixed"),
        ]
    )

    assert res.to_pydict() == {
        "asc": [[1, 3], [2, 4, 6], [3], [6, 11], None, [], [1, 3, None], None],
        "desc": [[3, 1], [6, 4, 2], [3], [11, 6], None, [], [None, 3, 1], None],
        "mixed": [[3, 1], [2, 4, 6], [3], [11, 6], None, [], [1, 3, None], None],
    }


def test_list_sort_fixed_size():
    table = MicroPartition.from_pydict(
        {
            "a": [[1, 3], [6, 2], [3, 3], [11, 6], None, [2, None], None],
            "b": [True, False, False, True, False, False, True],
        }
    ).eval_expression_list([col("a").cast(DataType.fixed_size_list(DataType.int64(), 2)), col("b")])

    res = table.eval_expression_list(
        [
            col("a").list.sort().alias("asc"),
            col("a").list.sort(True).alias("desc"),
            col("a").list.sort(col("b")).alias("mixed"),
        ]
    )

    assert res.to_pydict() == {
        "asc": [[1, 3], [2, 6], [3, 3], [6, 11], None, [2, None], None],
        "desc": [[3, 1], [6, 2], [3, 3], [11, 6], None, [None, 2], None],
        "mixed": [[3, 1], [2, 6], [3, 3], [11, 6], None, [2, None], None],
    }


def test_list_sort_with_groupby():
    df = daft.from_pydict({"group_col": [1, 1, 1, 2, 2, 2], "id_col": ["c", "a", "b", "e", "d", "a"]})

    # Group by group_col, aggregate id_col as a list.
    grouped_df = df.groupby("group_col").agg(daft.col("id_col").agg_list().alias("ids_col"))

    # Sort the list.
    result = grouped_df.select(col("group_col"), col("ids_col").list.sort()).sort("group_col", desc=False)
    result_dict = result.to_pydict()
    expected = {"group_col": [1, 2], "ids_col": [["a", "b", "c"], ["a", "d", "e"]]}
    assert result_dict == expected

    # Cast to fixed size list and sort.
    result = grouped_df.select(
        col("group_col"), col("ids_col").cast(DataType.fixed_size_list(DataType.string(), 3)).list.sort()
    ).sort("group_col", desc=False)
    result_dict = result.to_pydict()
    expected = {"group_col": [1, 2], "ids_col": [["a", "b", "c"], ["a", "d", "e"]]}
    assert result_dict == expected



from __future__ import annotations

import daft
from daft.sql.sql import SQLCatalog


def test_nested_access():
    df = daft.from_pydict(
        {
            "json": ['{"a": 1, "b": {"c": 2}}', '{"a": 3, "b": {"c": 4}}', '{"a": 5, "b": {"c": 6}}'],
            "list": [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            "dict": [{"a": 1, "b": 2}, {"a": 3, "b": 4}, {"a": 5, "b": 6}],
        }
    )

    catalog = SQLCatalog({"test": df})

    actual = daft.sql(
        """
    select
        json_query(json, '.b.c') as json_b_c,
        list[1] as list_1,
        list[0:1] as list_slice,
        dict['a'] as dict_a,
        struct_get(dict, 'a') as dict_a_2,
        cast(list as int[3])[1] as fsl_1,
        cast(list as int[3])[0:1] as fsl_slice
    from test
    """,
        catalog,
    ).collect()

    expected = df.select(
        daft.col("json").json.query(".b.c").alias("json_b_c"),
        daft.col("list").list.get(1).alias("list_1"),
        daft.col("list").list.slice(0, 1).alias("list_slice"),
        daft.col("dict").struct.get("a").alias("dict_a"),
        daft.col("dict").struct.get("a").alias("dict_a_2"),
        daft.col("list").cast(daft.DataType.fixed_size_list(daft.DataType.int32(), 3)).list.get(1).alias("fsl_1"),
        daft.col("list")
        .cast(daft.DataType.fixed_size_list(daft.DataType.int32(), 3))
        .list.slice(0, 1)
        .alias("fsl_slice"),
    ).collect()

    assert actual.to_pydict() == expected.to_pydict()



from __future__ import annotations

import pyarrow as pa
import pytest
from pyarrow import parquet as pq

import daft

from ..conftest import minio_create_bucket


@pytest.mark.integration()
def test_minio_parquet_bulk_readback(minio_io_config):
    bucket_name = "data-engineering-prod"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        target_paths = [
            "s3://data-engineering-prod/Y/part-00000-51723f93-0ba2-42f1-a58f-154f0ed40f28.c000.snappy.parquet",
            "s3://data-engineering-prod/Z/part-00000-6d5c7cc6-3b4a-443e-a46a-ca9e080bda1b.c000.snappy.parquet",
        ]
        data = {"x": [1, 2, 3, 4]}
        pa_table = pa.Table.from_pydict(data)
        for path in target_paths:
            pq.write_table(pa_table, path, filesystem=fs)

        readback = daft.recordbatch.read_parquet_into_pyarrow_bulk(target_paths, io_config=minio_io_config)
        assert len(readback) == len(target_paths)
        for tab in readback:
            assert tab.to_pydict() == data


@pytest.mark.integration()
def test_minio_parquet_read_no_files(minio_io_config):
    bucket_name = "data-engineering-prod"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        fs.touch("s3://data-engineering-prod/foo/file.txt")

        with pytest.raises(FileNotFoundError, match="Glob path had no matches:"):
            # Need to have a special character within the test path to trigger the matching logic
            daft.read_parquet(
                "s3://data-engineering-prod/foo/this-should-not-match-anything-and-this-file-should-not-exist-*.parquet",
                io_config=minio_io_config,
            )


@pytest.mark.integration()
def test_minio_parquet_ignore_marker_files(minio_io_config):
    bucket_name = "data-engineering-prod"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        target_paths = [
            "s3://data-engineering-prod/X/no_ext_parquet_metadata",
            "s3://data-engineering-prod/Y/part-00000-51723f93-0ba2-42f1-a58f-154f0ed40f28.c000.snappy.parquet",
            "s3://data-engineering-prod/Z/part-00000-6d5c7cc6-3b4a-443e-a46a-ca9e080bda1b.c000.snappy.parquet",
        ]
        data = {"x": [1, 2, 3, 4]}
        pa_table = pa.Table.from_pydict(data)
        for path in target_paths:
            pq.write_table(pa_table, path, filesystem=fs)

        marker_files = ["_metadata", "_SUCCESS", "_common_metadata", "a.crc"]
        for marker in marker_files:
            fs.touch(f"s3://{bucket_name}/X/{marker}")
            fs.touch(f"s3://{bucket_name}/Y/{marker}")
            fs.touch(f"s3://{bucket_name}/Z/{marker}")

        read = daft.read_parquet(f"s3://{bucket_name}/**", io_config=minio_io_config)
        assert read.to_pydict() == {"x": [1, 2, 3, 4] * 3}


@pytest.mark.integration()
def test_minio_parquet_ignore_marker_prefixes(minio_io_config):
    from datetime import datetime

    bucket_name = "data-engineering-prod"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        target_paths = [
            "s3://data-engineering-prod/X/part-00000-51723f93-0ba2-42f1-a58f-154f0ed40f28.c000.snappy.parquet",
            "s3://data-engineering-prod/Y/part-00000-6d5c7cc6-3b4a-443e-a46a-ca9e080bda1b.c000.snappy.parquet",
        ]
        data = {"x": [1, 2, 3, 4]}
        pa_table = pa.Table.from_pydict(data)
        for path in target_paths:
            pq.write_table(pa_table, path, filesystem=fs)

        marker_prefixes = ["_started", "_committed"]
        for marker_prefix in marker_prefixes:
            fs.touch(f"s3://{bucket_name}/X/{marker_prefix}_{datetime.now().isoformat()}")
            fs.touch(f"s3://{bucket_name}/Y/{marker_prefix}_{datetime.now().isoformat()}")

        read = daft.read_parquet(f"s3://{bucket_name}/**", io_config=minio_io_config)
        assert read.to_pydict() == {"x": [1, 2, 3, 4] * 2}


@pytest.mark.integration()
def test_minio_parquet_read_mismatched_schemas_no_pushdown(minio_io_config):
    # When we read files, we infer schema from the first file
    # Then when we read subsequent files, we want to be able to read the data still but add nulls for columns
    # that don't exist
    bucket_name = "data-engineering-prod"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        data_0 = pa.Table.from_pydict({"x": [1, 2, 3, 4]})
        pq.write_table(data_0, f"s3://{bucket_name}/data_0.parquet", filesystem=fs)
        data_1 = pa.Table.from_pydict({"y": [1, 2, 3, 4]})
        pq.write_table(data_1, f"s3://{bucket_name}/data_1.parquet", filesystem=fs)

        df = daft.read_parquet(
            [f"s3://{bucket_name}/data_0.parquet", f"s3://{bucket_name}/data_1.parquet"], io_config=minio_io_config
        )
        assert df.schema().column_names() == ["x"]
        assert df.to_pydict() == {"x": [1, 2, 3, 4, None, None, None, None]}


@pytest.mark.integration()
def test_minio_parquet_read_mismatched_schemas_with_pushdown(minio_io_config):
    # When we read files, we infer schema from the first file
    # Then when we read subsequent files, we want to be able to read the data still but add nulls for columns
    # that don't exist
    bucket_name = "data-engineering-prod"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        data_0 = pa.Table.from_pydict(
            {
                "x": [1, 2, 3, 4],
                "y": [1, 2, 3, 4],
                # NOTE: Need a column z here because Daft doesn't do a pushdown otherwise.
                "z": [1, 1, 1, 1],
            }
        )
        pq.write_table(data_0, f"s3://{bucket_name}/data_0.parquet", filesystem=fs)
        data_1 = pa.Table.from_pydict({"x": [5, 6, 7, 8]})
        pq.write_table(data_1, f"s3://{bucket_name}/data_1.parquet", filesystem=fs)

        df = daft.read_parquet(
            [f"s3://{bucket_name}/data_0.parquet", f"s3://{bucket_name}/data_1.parquet"], io_config=minio_io_config
        )
        df = df.select("x", "y")  # Applies column selection pushdown on each read
        assert df.schema().column_names() == ["x", "y"]
        assert df.to_pydict() == {"x": [1, 2, 3, 4, 5, 6, 7, 8], "y": [1, 2, 3, 4, None, None, None, None]}


@pytest.mark.integration()
def test_minio_parquet_read_mismatched_schemas_with_pushdown_no_rows_read(minio_io_config):
    # When we read files, we infer schema from the first file
    # Then when we read subsequent files, we want to be able to read the data still but add nulls for columns
    # that don't exist
    bucket_name = "data-engineering-prod"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        data_0 = pa.Table.from_pydict(
            {
                "x": [1, 2, 3, 4],
                # NOTE: Need a column z here because Daft doesn't do a pushdown otherwise.
                "z": [1, 1, 1, 1],
            }
        )
        pq.write_table(data_0, f"s3://{bucket_name}/data_0.parquet", filesystem=fs)
        data_1 = pa.Table.from_pydict({"y": [1, 2, 3, 4]})
        pq.write_table(data_1, f"s3://{bucket_name}/data_1.parquet", filesystem=fs)

        df = daft.read_parquet(
            [f"s3://{bucket_name}/data_0.parquet", f"s3://{bucket_name}/data_1.parquet"], io_config=minio_io_config
        )
        df = df.select("x")  # Applies column selection pushdown on each read
        assert df.schema().column_names() == ["x"]
        assert df.to_pydict() == {"x": [1, 2, 3, 4, None, None, None, None]}



from __future__ import annotations

import datetime

import pytest

from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_struct_get():
    table = MicroPartition.from_pydict(
        {
            "col": [
                {"foo": 1, "bar": "a"},
                {"foo": None, "bar": "b"},
                None,
                {"foo": 4, "bar": None},
            ]
        }
    )

    result = table.eval_expression_list([col("col").struct.get("foo"), col("col").struct.get("bar")])

    assert result.to_pydict() == {"foo": [1, None, None, 4], "bar": ["a", "b", None, None]}


def test_struct_get_logical_type():
    table = MicroPartition.from_pydict(
        {
            "col": [
                {"foo": datetime.date(2022, 1, 1)},
                {"foo": datetime.date(2022, 1, 2)},
                {"foo": None},
                None,
            ]
        }
    )

    result = table.eval_expression_list([col("col").struct.get("foo")])

    assert result.to_pydict() == {"foo": [datetime.date(2022, 1, 1), datetime.date(2022, 1, 2), None, None]}


def test_struct_get_bad_field():
    table = MicroPartition.from_pydict(
        {
            "col1": [
                {"foo": 1},
                {"foo": 2},
                {"foo": 3},
            ],
            "bar": ["a", "b", "c"],
        }
    )

    with pytest.raises(ValueError):
        table.eval_expression_list([col("col").struct.get("bar")])

    with pytest.raises(ValueError):
        table.eval_expression_list([col("bar").struct.get("foo")])



from __future__ import annotations

import operator as ops

import pytest

from daft.datatype import DataType
from daft.expressions import col
from tests.expressions.typing.conftest import (
    assert_typing_resolve_vs_runtime_behavior,
    has_supertype,
    is_comparable,
    is_numeric,
)


def comparable_type_validation(lhs: DataType, rhs: DataType) -> bool:
    return (
        is_comparable(lhs)
        and is_comparable(rhs)
        and has_supertype(lhs, rhs)
        and not ((is_numeric(lhs) and rhs == DataType.string()) or (is_numeric(rhs) and lhs == DataType.string()))
    )


@pytest.mark.parametrize("op", [ops.eq, ops.ne, ops.lt, ops.le, ops.gt, ops.ge])
def test_comparable(binary_data_fixture, op):
    lhs, rhs = binary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=binary_data_fixture,
        expr=op(col(lhs.name()), col(rhs.name())),
        run_kernel=lambda: op(lhs, rhs),
        resolvable=comparable_type_validation(lhs.datatype(), rhs.datatype()),
    )



from __future__ import annotations

from .functions import (
    monotonically_increasing_id,
    columns_sum,
    columns_mean,
    columns_avg,
    columns_min,
    columns_max,
    row_number,
    rank,
    dense_rank,
)
from .llm_generate import llm_generate

__all__ = [
    "columns_avg",
    "columns_max",
    "columns_mean",
    "columns_min",
    "columns_sum",
    "dense_rank",
    "llm_generate",
    "monotonically_increasing_id",
    "rank",
    "row_number",
]



from __future__ import annotations

import pyarrow as pa
import pytest

import daft
from daft import DataType
from daft.logical.schema import Schema
from daft.recordbatch import MicroPartition, recordbatch_io
from tests.recordbatch.recordbatch_io.test_parquet import _parquet_write_helper


@pytest.mark.parametrize(
    ["data", "schema", "expected"],
    [
        # Test that a cast occurs (in this case, int64 -> int8)
        (
            pa.Table.from_pydict({"foo": pa.array([1, 2, 3], type=pa.int64())}),
            Schema._from_field_name_and_types([("foo", DataType.int8())]),
            MicroPartition.from_pydict({"foo": daft.Series.from_arrow(pa.array([1, 2, 3], type=pa.int8()))}),
        ),
        # Test what happens if a cast should occur, but fails at runtime (in this case, a potentially bad cast from utf8->int64)
        (
            pa.Table.from_pydict({"foo": pa.array(["1", "2", "FAIL"], type=pa.string())}),
            Schema._from_field_name_and_types([("foo", DataType.int64())]),
            # NOTE: cast failures will become a Null value
            MicroPartition.from_pydict({"foo": daft.Series.from_arrow(pa.array([1, 2, None], type=pa.int64()))}),
        ),
        # Test reordering of columns
        (
            pa.Table.from_pydict({"foo": pa.array([1, 2, 3]), "bar": pa.array([4, 5, 6])}),
            Schema._from_field_name_and_types([("bar", DataType.int64()), ("foo", DataType.int64())]),
            MicroPartition.from_pydict({"bar": pa.array([4, 5, 6]), "foo": pa.array([1, 2, 3])}),
        ),
        # Test automatic insertion of null values for missing column
        (
            pa.Table.from_pydict({"foo": pa.array([1, 2, 3])}),
            Schema._from_field_name_and_types([("bar", DataType.int64()), ("foo", DataType.int64())]),
            MicroPartition.from_pydict(
                {"bar": pa.array([None, None, None], type=pa.int64()), "foo": pa.array([1, 2, 3])}
            ),
        ),
    ],
)
def test_parquet_cast_at_read_time(data, schema, expected):
    with _parquet_write_helper(data) as f:
        table = recordbatch_io.read_parquet(f, schema)
        assert table.schema() == schema
        assert table.to_arrow() == expected.to_arrow()



from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable

from daft.daft import (
    LocalPhysicalPlan,
    NativeExecutor,
    PyDaftExecutionConfig,
    RayPartitionRef,
    RaySwordfishTask,
)
from daft.recordbatch.micropartition import MicroPartition
from daft.runners.constants import MAX_SWORDFISH_ACTOR_RESTARTS, MAX_SWORDFISH_ACTOR_TASK_RETRIES
from daft.runners.partitioning import PartitionMetadata

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

try:
    import ray
    import ray.util.scheduling_strategies
except ImportError:
    raise

logger = logging.getLogger(__name__)


@ray.remote(max_restarts=MAX_SWORDFISH_ACTOR_RESTARTS, max_task_retries=MAX_SWORDFISH_ACTOR_TASK_RETRIES)
class RaySwordfishWorker:
    """RaySwordfishWorker is a ray actor that runs local physical plans on swordfish.

    It is a stateless, async actor, and can run multiple plans concurrently and is able to retry itself and it's tasks.
    """

    def __init__(self, daft_execution_config: PyDaftExecutionConfig) -> None:
        self.native_executor = NativeExecutor()
        self.daft_execution_config = daft_execution_config

    async def run_plan(
        self,
        plan: LocalPhysicalPlan,
        psets: dict[str, list[ray.ObjectRef]],
    ) -> AsyncGenerator[MicroPartition, None]:
        """Run a plan on swordfish and yield partitions."""
        psets_gathered = {k: await asyncio.gather(*v) for k, v in psets.items()}
        psets_mp = {k: [v._micropartition for v in v] for k, v in psets_gathered.items()}
        async for partition in self.native_executor.run_async(plan, psets_mp, self.daft_execution_config, None):
            if partition is None:
                break
            mp = MicroPartition._from_pymicropartition(partition)
            yield mp

    def concat_and_get_metadata(self, *partitions: MicroPartition) -> tuple[PartitionMetadata, MicroPartition]:
        """Concatenate a list of partitions and return the metadata and the concatenated partition."""
        concated = MicroPartition.concat(list(partitions))
        return PartitionMetadata.from_table(concated), concated


@dataclass
class RaySwordfishTaskHandle:
    """RaySwordfishTaskHandle is a handle to a task that is running on a swordfish actor.

    It is used to asynchronously get the result of the task, cancel the task, and perform any post-task cleanup.
    """

    result_handle: ray.ObjectRef
    actor_handle: ray.actor.ActorHandle
    done_callback: Callable[[asyncio.Task[RayPartitionRef]], None]
    task_memory_cost: int

    async def _get_result(self) -> RayPartitionRef:
        results = []
        async for result in self.result_handle:
            results.append(result)

        metadata, result = self.actor_handle.concat_and_get_metadata.options(num_returns=2).remote(*results)
        metadata = await metadata
        return RayPartitionRef(result, metadata.num_rows, metadata.size_bytes)

    async def get_result(self) -> RayPartitionRef:
        task = asyncio.create_task(self._get_result())
        task.add_done_callback(self.done_callback)
        return await task

    def cancel(self) -> None:
        ray.cancel(self.result_handle)


class RaySwordfishWorkerHandle:
    """RaySwordfishWorkerHandle is a wrapper around a ray swordfish actor.

    It is used to submit tasks to the worker and keep track of the worker's node id, handle, number of cpus, total and available memory.
    """

    def __init__(
        self,
        actor_node_id: str,
        actor_handle: ray.actor.ActorHandle,
        num_cpus: int,
        total_memory_bytes: int,
    ) -> None:
        self.actor_node_id = actor_node_id
        self.actor_handle = actor_handle
        self.num_cpus = num_cpus
        self.total_memory_bytes = total_memory_bytes
        self.available_memory_bytes = total_memory_bytes

    def get_actor_node_id(self) -> str:
        return self.actor_node_id

    def get_num_cpus(self) -> int:
        return self.num_cpus

    def get_total_memory_bytes(self) -> int:
        return self.total_memory_bytes

    def get_available_memory_bytes(self) -> int:
        return self.available_memory_bytes

    def add_back_memory(self, memory_bytes: int) -> None:
        self.available_memory_bytes += memory_bytes

    def submit_task(self, task: RaySwordfishTask) -> RaySwordfishTaskHandle:
        self.available_memory_bytes -= task.estimated_memory_cost()
        psets = {k: [v.object_ref for v in v] for k, v in task.psets().items()}
        memory_to_return = task.estimated_memory_cost()

        def done_callback(_result: asyncio.Task[RayPartitionRef]) -> None:
            self.add_back_memory(memory_to_return)

        return RaySwordfishTaskHandle(
            self.actor_handle.run_plan.remote(task.plan(), psets),
            self.actor_handle,
            done_callback,
            task.estimated_memory_cost(),
        )

    def shutdown(self) -> None:
        ray.kill(self.actor_handle)


class RaySwordfishWorkerManager:
    """RaySwordfishWorkerManager is keeps track of all swordfish workers, creates new workers, and shuts down workers.

    It should have a global view of all workers and their resources.
    It is also responsible for autoscaling the number of workers.
    """

    def __init__(self, daft_execution_config: PyDaftExecutionConfig) -> None:
        self.workers: dict[str, RaySwordfishWorkerHandle] = {}
        self.daft_execution_config = daft_execution_config
        self._initialize_workers()

    def _initialize_workers(self) -> None:
        for node in ray.nodes():
            if (
                "Resources" in node
                and "CPU" in node["Resources"]
                and "memory" in node["Resources"]
                and node["Resources"]["CPU"] > 0
                and node["Resources"]["memory"] > 0
            ):
                actor = RaySwordfishWorker.options(  # type: ignore[attr-defined]
                    num_cpus=node["Resources"]["CPU"],
                    scheduling_strategy=ray.util.scheduling_strategies.NodeAffinitySchedulingStrategy(
                        node_id=node["NodeID"],
                        soft=True,
                    ),
                ).remote(self.daft_execution_config)
                self.workers[node["NodeID"]] = RaySwordfishWorkerHandle(
                    node["NodeID"],
                    actor,
                    int(node["Resources"]["CPU"]),
                    int(node["Resources"]["memory"]),
                )

    def submit_task_to_worker(self, task: RaySwordfishTask, worker_id: str) -> RaySwordfishTaskHandle:
        return self.workers[worker_id].submit_task(task)

    def get_worker_resources(self) -> list[tuple[str, int, int]]:
        return [
            (
                worker.get_actor_node_id(),
                worker.get_num_cpus(),
                worker.get_available_memory_bytes(),
            )
            for worker in self.workers.values()
        ]

    def try_autoscale(self, num_workers: int) -> None:
        """Try to autoscale the number of workers. this is a hint to the autoscaler to add more workers if needed."""
        ray.autoscaler.sdk.request_resources(
            num_cpus=num_workers,
        )

    def shutdown(self) -> None:
        for worker in self.workers.values():
            worker.shutdown()



from __future__ import annotations

import pandas as pd
import pyarrow as pa
import pytest

from daft import col


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize("percentiles_expected", [(0.5, [2.0]), ([0.5], [[2.0]]), ([0.5, 0.5], [[2.0, 2.0]])])
def test_approx_percentiles_global(make_df, repartition_nparts, percentiles_expected):
    percentiles, expected = percentiles_expected

    daft_df = make_df(
        {
            "id": [1, 2, 3],
            "values": [1, 2, 3],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.agg(
        [
            col("values").approx_percentiles(percentiles).alias("percentiles"),
        ]
    )
    daft_cols = daft_df.to_pydict()
    pd.testing.assert_series_equal(
        pd.Series(daft_cols["percentiles"]), pd.Series(expected), check_exact=False, rtol=0.02
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
@pytest.mark.parametrize("percentiles_expected", [(0.5, [2.0]), ([0.5], [[2.0]]), ([0.5, 0.5], [[2.0, 2.0]])])
def test_approx_percentiles_global_with_nulls(make_df, repartition_nparts, percentiles_expected):
    percentiles, expected = percentiles_expected
    daft_df = make_df(
        {
            "id": [1, 2, 3, 4],
            "values": [1, None, 2, 3],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.agg(
        [
            col("values").approx_percentiles(percentiles).alias("percentiles"),
        ]
    )
    daft_cols = daft_df.to_pydict()
    pd.testing.assert_series_equal(
        pd.Series(daft_cols["percentiles"]), pd.Series(expected), check_exact=False, rtol=0.02
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize("percentiles_expected", [(0.5, [None]), ([0.5], [None]), ([0.5, 0.5], [None])])
def test_approx_percentiles_global_all_nulls(make_df, repartition_nparts, percentiles_expected):
    percentiles, expected = percentiles_expected
    daft_df = make_df(
        {
            "id": [1, 2, 3],
            "values": pa.array([None, None, None], type=pa.int64()),
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.agg(
        [
            col("values").approx_percentiles(percentiles).alias("percentiles"),
        ]
    )
    daft_cols = daft_df.to_pydict()
    pd.testing.assert_series_equal(pd.Series(daft_cols["percentiles"]), pd.Series(expected))


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
@pytest.mark.parametrize(
    "percentiles_expected", [(0.5, [2.0, 2.0]), ([0.5], [[2.0], [2.0]]), ([0.5, 0.5], [[2.0, 2.0], [2.0, 2.0]])]
)
def test_approx_percentiles_groupby(make_df, repartition_nparts, percentiles_expected):
    percentiles, expected = percentiles_expected
    daft_df = make_df(
        {
            "id": [1, 1, 1, 2],
            "values": [1, 2, 3, 2],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.groupby("id").agg(
        [
            col("values").approx_percentiles(percentiles).alias("percentiles"),
        ]
    )
    daft_cols = daft_df.to_pydict()
    pd.testing.assert_series_equal(
        pd.Series(daft_cols["percentiles"]), pd.Series(expected), check_exact=False, rtol=0.02
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2, 7])
@pytest.mark.parametrize(
    "percentiles_expected",
    [(0.5, [2.0, 2.0, None]), ([0.5], [[2.0], [2.0], None]), ([0.5, 0.5], [[2.0, 2.0], [2.0, 2.0], None])],
)
def test_approx_percentiles_groupby_with_nulls(make_df, repartition_nparts, percentiles_expected):
    percentiles, expected = percentiles_expected
    daft_df = make_df(
        {
            "id": [1, 1, 1, 2, 2, 3],
            "values": [1, 2, 3, 2, None, None],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.groupby("id").agg(
        [
            col("values").approx_percentiles(percentiles).alias("percentiles"),
        ]
    )
    daft_cols = daft_df.to_pandas()
    pd.testing.assert_series_equal(
        daft_cols.sort_values("id")["percentiles"],
        pd.Series(expected),
        check_exact=False,
        rtol=0.02,
        check_index=False,
        check_names=False,
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize(
    "percentiles_expected", [(0.5, [None, None]), ([0.5], [None, None]), ([0.5, 0.5], [None, None])]
)
def test_approx_percentiles_groupby_all_nulls(make_df, repartition_nparts, percentiles_expected):
    percentiles, expected = percentiles_expected
    daft_df = make_df(
        {
            "id": [1, 1, 2],
            "values": pa.array([None, None, None], type=pa.int64()),
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.groupby("id").agg(
        [
            col("values").approx_percentiles(percentiles).alias("percentiles"),
        ]
    )
    daft_cols = daft_df.to_pydict()
    assert daft_cols["percentiles"] == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
@pytest.mark.parametrize(
    "percentiles_expected",
    [
        (0.5, [2.0, 2.0]),
        ([0.5], [[2.0], [2.0]]),
        ([0.5, 0.5], [[2.0, 2.0], [2.0, 2.0]]),
    ],
)
def test_approx_percentiles_groupby_with_alias(make_df, repartition_nparts, percentiles_expected):
    percentiles, expected = percentiles_expected
    daft_df = make_df(
        {
            "id": [1, 1, 1, 2],
            "values": [1, 2, 3, 2],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.groupby(daft_df["id"].alias("id_alias")).agg(
        [
            col("values").approx_percentiles(percentiles).alias("percentiles"),
        ]
    )
    daft_cols = daft_df.to_pydict()
    pd.testing.assert_series_equal(
        pd.Series(daft_cols["percentiles"]),
        pd.Series(expected),
        check_exact=False,
        rtol=0.02,
    )



from __future__ import annotations

import pyarrow as pa
import pytest

from daft.series import Series


@pytest.mark.parametrize(
    "input,fill_value,expected",
    [
        # No broadcast
        [[1, 2, None], [3, 3, 3], [1, 2, 3]],
        # Broadcast input
        [[None], [3, 3, 3], [3, 3, 3]],
        # Broadcast fill_value
        [[1, 2, None], [3], [1, 2, 3]],
        # Empty
        [[], [], []],
    ],
)
def test_series_fill_null(input, fill_value, expected) -> None:
    s = Series.from_arrow(pa.array(input, pa.int64()))
    fill_value = Series.from_arrow(pa.array(fill_value, pa.int64()))
    filled = s.fill_null(fill_value)
    assert filled.to_pylist() == expected


def test_series_fill_null_bad_input() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3], pa.int64()))
    with pytest.raises(ValueError, match="expected another Series but got"):
        s.fill_null([1, 2, 3])



from __future__ import annotations

import decimal
from datetime import date, datetime, time

import numpy as np
import pytest
import pytz
import xxhash

from daft.datatype import DataType
from daft.series import Series
from tests.test_datatypes import daft_numeric_types


@pytest.mark.parametrize(
    "nbytes, dtype",
    [
        (1, DataType.uint8()),
        (1, DataType.int8()),
        (2, DataType.uint16()),
        (2, DataType.int16()),
        (4, DataType.uint32()),
        (4, DataType.int32()),
        (8, DataType.uint64()),
        (8, DataType.int64()),
    ],
)
def test_hash_int_array_with_reference(nbytes, dtype):
    arr = Series.from_numpy(np.random.randint(0, 127, 100)).cast(dtype)

    hashed = arr.hash()
    for v, hv in zip(arr.to_pylist(), hashed.to_pylist()):
        vbytes = v.to_bytes(nbytes, "little")
        ref_value = xxhash.xxh3_64_intdigest(vbytes)
        assert hv == ref_value

    hashed_again = arr.hash(hashed)

    for v, hashed_value, hashed_again_value in zip(arr.to_pylist(), hashed.to_pylist(), hashed_again.to_pylist()):
        vbytes = v.to_bytes(nbytes, "little")
        ref_value = xxhash.xxh3_64_intdigest(vbytes, seed=hashed_value)
        assert hashed_again_value == ref_value


def test_hash_bool_array_with_reference():
    arr = Series.from_pylist([False, True, True])
    expected = [xxhash.xxh3_64_intdigest(b"0"), xxhash.xxh3_64_intdigest(b"1"), xxhash.xxh3_64_intdigest(b"1")]
    hashed = arr.hash()
    assert hashed.to_pylist() == expected

    hashed_again = arr.hash(hashed)
    expected = [
        xxhash.xxh3_64_intdigest(b"0", expected[0]),
        xxhash.xxh3_64_intdigest(b"1", expected[1]),
        xxhash.xxh3_64_intdigest(b"1", expected[2]),
    ]
    assert hashed_again.to_pylist() == expected


def test_hash_str_array_with_reference():
    arr = Series.from_pylist(["hi", "bye", None])
    expected = [xxhash.xxh3_64_intdigest(b"hi"), xxhash.xxh3_64_intdigest(b"bye"), xxhash.xxh3_64_intdigest(b"")]
    hashed = arr.hash()
    assert hashed.to_pylist() == expected

    hashed_again = arr.hash(hashed)
    expected = [
        xxhash.xxh3_64_intdigest(b"hi", expected[0]),
        xxhash.xxh3_64_intdigest(b"bye", expected[1]),
        xxhash.xxh3_64_intdigest(b"", expected[2]),
    ]
    assert hashed_again.to_pylist() == expected


def test_hash_binary_array_with_reference():
    arr = Series.from_pylist([b"hi", b"bye", None])
    expected = [xxhash.xxh3_64_intdigest(b"hi"), xxhash.xxh3_64_intdigest(b"bye"), xxhash.xxh3_64_intdigest(b"")]
    hashed = arr.hash()
    assert hashed.to_pylist() == expected

    hashed_again = arr.hash(hashed)
    expected = [
        xxhash.xxh3_64_intdigest(b"hi", expected[0]),
        xxhash.xxh3_64_intdigest(b"bye", expected[1]),
        xxhash.xxh3_64_intdigest(b"", expected[2]),
    ]
    assert hashed_again.to_pylist() == expected


def test_hash_fixed_size_binary_array_with_reference():
    import pyarrow as pa

    arr = Series.from_arrow(pa.array([b"foo", b"bar", None], type=pa.binary(3)))
    expected = [
        xxhash.xxh3_64_intdigest(b"foo"),
        xxhash.xxh3_64_intdigest(b"bar"),
        xxhash.xxh3_64_intdigest(b"\x00\x00\x00"),
    ]
    hashed = arr.hash()
    assert hashed.to_pylist() == expected

    hashed_again = arr.hash(hashed)
    expected = [
        xxhash.xxh3_64_intdigest(b"foo", expected[0]),
        xxhash.xxh3_64_intdigest(b"bar", expected[1]),
        xxhash.xxh3_64_intdigest(b"\x00\x00\x00", expected[2]),
    ]
    assert hashed_again.to_pylist() == expected


def test_hash_null_array_with_reference():
    arr = Series.from_pylist([None, None, None])
    expected = [xxhash.xxh3_64_intdigest(b""), xxhash.xxh3_64_intdigest(b""), xxhash.xxh3_64_intdigest(b"")]
    hashed = arr.hash()
    assert hashed.to_pylist() == expected

    hashed_again = arr.hash(hashed)
    expected = [
        xxhash.xxh3_64_intdigest(b"", expected[0]),
        xxhash.xxh3_64_intdigest(b"", expected[1]),
        xxhash.xxh3_64_intdigest(b"", expected[2]),
    ]
    assert hashed_again.to_pylist() == expected


def test_hash_int_array_with_bad_seed():
    arr = Series.from_numpy(np.random.randint(0, 127, 100)).cast(DataType.uint64())

    bad_seed = Series.from_numpy(np.random.randint(0, 127, 100)).cast(DataType.float64())

    with pytest.raises(ValueError, match="We can only use UInt64 as a seed"):
        arr.hash(bad_seed)


def test_hash_int_array_with_bad_length():
    arr = Series.from_numpy(np.random.randint(0, 127, 100)).cast(DataType.uint64())

    bad_seed = Series.from_numpy(np.random.randint(0, 127, 99)).cast(DataType.uint64())

    with pytest.raises(ValueError, match="seed length does not match array length"):
        arr.hash(bad_seed)


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_hash_list_array_no_seed(dtype):
    arr = Series.from_pylist([[1, 2], [1, 3], [1, 2], [1, 2, 3], [], [], [2, 1]]).cast(DataType.list(dtype))

    hashed = arr.hash().to_pylist()
    assert hashed[0] == hashed[2]
    assert hashed[4] == hashed[5]

    different_inds = [0, 1, 3, 4, 6]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
@pytest.mark.parametrize("seed", [1, 2, 42])
def test_hash_list_array_seeded(dtype, seed):
    arr = Series.from_pylist([[1, 2], [1, 3], [1, 2], [1, 2, 3], [], [], [2, 1]]).cast(DataType.list(dtype))
    seeds = Series.from_pylist([seed] * 9).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()
    assert hashed[0] == hashed[2]
    assert hashed[4] == hashed[5]

    different_inds = [0, 1, 3, 4, 6]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_hash_list_array_no_seed_with_invalid(dtype):
    arr = Series.from_pylist([[1, 2], [1, 3], [1, 2], [1, 2, 3], [], [], None, [2, 1], None]).cast(DataType.list(dtype))

    hashed = arr.hash().to_pylist()
    assert hashed[0] == hashed[2]
    assert hashed[4] == hashed[5]
    assert hashed[6] is None
    assert hashed[8] is None

    different_inds = [0, 1, 3, 4, 7]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
@pytest.mark.parametrize("seed", [1, 2, 42])
def test_hash_list_array_seeded_with_invalid(dtype, seed):
    arr = Series.from_pylist([[1, 2], [1, 3], [1, 2], [1, 2, 3], [], [], None, [2, 1], None]).cast(DataType.list(dtype))
    seeds = Series.from_pylist([seed] * 9).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()
    assert hashed[0] == hashed[2]
    assert hashed[4] == hashed[5]
    assert hashed[6] is None
    assert hashed[8] is None

    different_inds = [0, 1, 3, 4, 7]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_hash_list_array_different_seeds(dtype):
    arr = Series.from_pylist([[1, 2], [1, 2], [1, 2], [1, 2]]).cast(DataType.list(dtype))
    seeds = Series.from_pylist([1, 2, 3, 4]).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()

    different_inds = [0, 1, 2, 3]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_hash_list_array_nested_lists(dtype):
    arr = Series.from_pylist(
        [
            [[1, 2], [3, 4]],
            [[1, 2], [3, 5]],
            [[1, 2], [3, 4]],
            [[3, 4], [1, 2]],
            [[1], [2]],
            [[], []],
            [[]],
            [],
            [[1, 2, 3]],
            [[1], [2], [3]],
        ]
    ).cast(DataType.list(DataType.list(dtype)))

    hashed = arr.hash().to_pylist()
    assert hashed[0] == hashed[2]

    different_inds = [0, 1, 3, 4, 5, 6, 7, 8, 9]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_hash_list_array_consistency(dtype):
    data = [[1, 2], [1, 3], [1, 2], [1, 2, 3], [], [], None, [2, 1], None]
    arr1 = Series.from_pylist(data).cast(DataType.list(dtype))
    arr2 = Series.from_pylist(data).cast(DataType.list(dtype))

    hashed1 = arr1.hash().to_pylist()
    hashed2 = arr2.hash().to_pylist()
    assert hashed1 == hashed2


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_hash_fixed_size_list_array_no_seed(dtype):
    arr = Series.from_pylist([[1, 2], [1, 3], [1, 2], [1, 4], [5, 5], [5, 5], [2, 1]]).cast(
        DataType.fixed_size_list(dtype, 2)
    )

    hashed = arr.hash().to_pylist()
    assert hashed[0] == hashed[2]
    assert hashed[4] == hashed[5]

    different_inds = [0, 1, 3, 4, 6]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
@pytest.mark.parametrize("seed", [1, 2, 42])
def test_hash_fixed_size_list_array_seeded(dtype, seed):
    arr = Series.from_pylist([[1, 2], [1, 3], [1, 2], [1, 4], [5, 5], [5, 5], [2, 1]]).cast(
        DataType.fixed_size_list(dtype, 2)
    )
    seeds = Series.from_pylist([seed] * 9).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()
    assert hashed[0] == hashed[2]
    assert hashed[4] == hashed[5]

    different_inds = [0, 1, 3, 4, 6]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_hash_fixed_size_list_array_no_seed_with_invalid(dtype):
    arr = Series.from_pylist([[1, 2], [1, 3], [1, 2], [1, 4], [5, 5], [5, 5], None, [2, 1], None]).cast(
        DataType.fixed_size_list(dtype, 2)
    )

    hashed = arr.hash().to_pylist()
    assert hashed[0] == hashed[2]
    assert hashed[4] == hashed[5]
    assert hashed[6] is None
    assert hashed[8] is None

    different_inds = [0, 1, 3, 4, 7]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
@pytest.mark.parametrize("seed", [1, 2, 42])
def test_hash_fixed_size_list_array_seeded_with_invalid(dtype, seed):
    arr = Series.from_pylist([[1, 2], [1, 3], [1, 2], [1, 4], [5, 5], [5, 5], None, [2, 1], None]).cast(
        DataType.fixed_size_list(dtype, 2)
    )
    seeds = Series.from_pylist([seed] * 9).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()
    assert hashed[0] == hashed[2]
    assert hashed[4] == hashed[5]
    assert hashed[6] is None
    assert hashed[8] is None

    different_inds = [0, 1, 3, 4, 7]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_hash_fixed_size_list_array_different_seeds(dtype):
    arr = Series.from_pylist([[1, 2], [1, 2], [1, 2], [1, 2]]).cast(DataType.fixed_size_list(dtype, 2))
    seeds = Series.from_pylist([1, 2, 3, 4]).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()

    different_inds = [0, 1, 2, 3]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_hash_fixed_size_list_array_consistency(dtype):
    data = [[1, 2], [1, 3], [1, 2], [1, 4], [5, 5], [5, 5], None, [2, 1], None]
    arr1 = Series.from_pylist(data).cast(DataType.fixed_size_list(dtype, 2))
    arr2 = Series.from_pylist(data).cast(DataType.fixed_size_list(dtype, 2))

    hashed1 = arr1.hash().to_pylist()
    hashed2 = arr2.hash().to_pylist()
    assert hashed1 == hashed2


@pytest.mark.parametrize("dtype", daft_numeric_types)
@pytest.mark.parametrize("seed", [None, 123])
def test_hash_struct(dtype, seed):
    data = [
        {"a": 1, "b": 2},
        {"a": 3, "b": 4},
        {"a": 1, "b": 4},
        {"a": 2, "b": 1},
        None,
        {"b": 2, "a": 1},
        {"b": 1, "a": 2},
        None,
    ]
    arr = Series.from_pylist(data).cast(DataType.struct({"a": dtype, "b": dtype}))

    seeds = None if seed is None else Series.from_pylist([seed] * len(data)).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()
    assert hashed[0] == hashed[5]
    assert hashed[3] == hashed[6]
    assert hashed[4] is None and hashed[-1] is None

    different_inds = [0, 1, 2, 3]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
@pytest.mark.parametrize("seed", [None, 123])
def test_hash_struct_nested(dtype, seed):
    data = [
        {"a": 1, "b": {"c": 1, "d": 2}},
        {"a": 3, "b": {"c": 3, "d": 4}},
        {"a": 1, "b": {"c": 3, "d": 4}},
        {"a": 2, "b": {"c": 1, "d": 2}},
        None,
        {"b": {"c": 1, "d": 2}, "a": 1},
        {"b": {"c": 1, "d": 2}, "a": 2},
        {"a": 1, "b": {"c": 2, "d": 1}},
        None,
    ]
    arr = Series.from_pylist(data).cast(DataType.struct({"a": dtype, "b": DataType.struct({"c": dtype, "d": dtype})}))

    seeds = None if seed is None else Series.from_pylist([seed] * len(data)).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()
    assert hashed[0] == hashed[5]
    assert hashed[3] == hashed[6]
    assert hashed[4] is None and hashed[-1] is None

    different_inds = [0, 1, 2, 3, 7]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
@pytest.mark.parametrize("seed", [None, 123])
def test_hash_struct_sublist(dtype, seed):
    data = [
        {"a": 1, "b": [2, 3, 4]},
        {"a": 3, "b": [5, 6, 7]},
        {"a": 1, "b": [5, 6, 7]},
        {"a": 2, "b": [2, 3, 4]},
        None,
        {"b": [2, 3, 4], "a": 1},
        {"b": [2, 3, 4], "a": 2},
        {"a": 1, "b": [2, 4, 3]},
        None,
    ]
    arr = Series.from_pylist(data).cast(DataType.struct({"a": dtype, "b": DataType.list(dtype)}))

    seeds = None if seed is None else Series.from_pylist([seed] * len(data)).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()
    assert hashed[0] == hashed[5]
    assert hashed[3] == hashed[6]
    assert hashed[4] is None and hashed[-1] is None

    different_inds = [0, 1, 2, 3, 7]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize("dtype", daft_numeric_types)
@pytest.mark.parametrize("seed", [None, 123])
def test_hash_struct_with_nones(dtype, seed):
    data = [
        {"a": 1, "b": 2},
        {"a": None, "b": 2},
        {"a": 0, "b": 2},
        {"a": None, "b": None},
        {"a": 0, "b": 0},
        None,
        {"a": None, "b": 2},
        None,
    ]
    arr = Series.from_pylist(data).cast(DataType.struct({"a": dtype, "b": dtype}))

    seeds = None if seed is None else Series.from_pylist([seed] * len(data)).cast(DataType.uint64())

    hashed = arr.hash(seeds).to_pylist()
    assert hashed[1] == hashed[6]
    assert hashed[5] is None and hashed[-1] is None

    different_inds = [0, 1, 2, 3, 4]
    for i in range(len(different_inds)):
        for j in range(i):
            assert hashed[different_inds[i]] != hashed[different_inds[j]]


@pytest.mark.parametrize(
    "dtype",
    [
        DataType.uint8(),
        DataType.uint16(),
        DataType.uint32(),
        DataType.uint64(),
        DataType.int8(),
        DataType.int16(),
        DataType.int32(),
        DataType.int64(),
    ],
)
def test_murmur3_32_hash_int(dtype):
    arr = Series.from_pylist([34, None]).cast(dtype)
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [2017239379, None]


@pytest.mark.parametrize(
    "dtype",
    [
        DataType.int8(),
        DataType.int16(),
        DataType.int32(),
        DataType.int64(),
    ],
)
def test_murmur3_32_hash_signed_int(dtype):
    arr = Series.from_pylist([-1, 34, None]).cast(dtype)
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [1651860712, 2017239379, None]


def test_murmur3_32_hash_string():
    arr = Series.from_pylist(["iceberg", None])
    assert arr.datatype() == DataType.string()
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [1210000089, None]


def test_murmur3_32_hash_bytes():
    arr = Series.from_pylist([b"\x00\x01\x02\x03", None])
    assert arr.datatype() == DataType.binary()
    hashes = arr.murmur3_32()
    java_answer = -188683207
    assert hashes.to_pylist() == [java_answer, None]


def test_murmur3_32_hash_fixed_sized_bytes():
    import pyarrow as pa

    arr = Series.from_arrow(pa.array([b"\x00\x01\x02\x03", None], type=pa.binary(4)))
    assert arr.datatype() == DataType.fixed_size_binary(4)
    hashes = arr.murmur3_32()
    java_answer = -188683207
    assert hashes.to_pylist() == [java_answer, None]


def test_murmur3_32_hash_date():
    arr = Series.from_pylist([date(2017, 11, 16), None])
    assert arr.datatype() == DataType.date()
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [-653330422, None]


def test_murmur3_32_hash_time():
    arr = Series.from_pylist([time(22, 31, 8, 0), None])
    assert arr.datatype() == DataType.time("us")
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [-662762989, None]


def test_murmur3_32_hash_time_nanoseconds():
    arr = Series.from_pylist([time(22, 31, 8, 0), None])
    arr = arr.cast(DataType.time("ns"))
    assert arr.datatype() == DataType.time("ns")
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [-662762989, None]


def test_murmur3_32_hash_timestamp():
    arr = Series.from_pylist([datetime(2017, 11, 16, 22, 31, 8), None])
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [-2047944441, None]


def test_murmur3_32_hash_timestamp_with_tz():
    dt = datetime(2017, 11, 16, 14, 31, 8)
    pst = pytz.timezone("US/Pacific")
    dt = pst.localize(dt)
    arr = Series.from_pylist([dt, None])
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [-2047944441, None]


def test_murmur3_32_hash_timestamp_with_tz_nanoseconds():
    dt = datetime(2017, 11, 16, 14, 31, 8)
    pst = pytz.timezone("US/Pacific")
    dt = pst.localize(dt)
    arr = Series.from_pylist([dt, None])
    arr = arr.cast(DataType.timestamp("ns", "UTC"))
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [-2047944441, None]


def test_murmur3_32_hash_decimal_unscaled():
    arr = Series.from_pylist([decimal.Decimal(1420), None])
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [-500754589, None]


def test_murmur3_32_hash_decimal_scaled():
    arr = Series.from_pylist([decimal.Decimal("14.20"), None])
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [-500754589, None]


def test_murmur3_32_hash_decimal_full_scaled():
    arr = Series.from_pylist([decimal.Decimal(".00001420"), None])
    hashes = arr.murmur3_32()
    assert hashes.to_pylist() == [-500754589, None]



from __future__ import annotations


def test_endswith(test_expression):
    test_data = ["hello", "world", "python"]
    expected = [False, False, True]
    test_expression(
        data=test_data, expected=expected, name="endswith", sql_name="ends_with", namespace="str", args=["thon"]
    )


def test_startswith(test_expression):
    test_data = ["hello", "world", "python"]
    expected = [True, False, False]
    test_expression(
        data=test_data, expected=expected, name="startswith", sql_name="starts_with", namespace="str", args=["hello"]
    )


def contains(test_expression):
    test_data = ["hello", "world", "python"]
    expected = [True, False, True]
    test_expression(
        data=test_data, expected=expected, name="contains", sql_name="contains", namespace="str", args=[["lo"]]
    )


def test_upper(test_expression):
    test_data = ["hello", "world", "python"]
    expected = ["HELLO", "WORLD", "PYTHON"]
    test_expression(
        data=test_data,
        expected=expected,
        name="upper",
        namespace="str",
    )


def test_upper_full_null(test_expression):
    test_data = [None, None, None]
    expected = [None, None, None]
    test_expression(
        data=test_data,
        expected=expected,
        name="upper",
        namespace="str",
    )


def test_capitalize(test_expression):
    test_data = ["hello", "world", "python"]
    expected = ["Hello", "World", "Python"]
    test_expression(
        data=test_data,
        expected=expected,
        name="capitalize",
        namespace="str",
    )


def test_lower(test_expression):
    test_data = ["HELLO", "WORLD", "PYTHON"]
    expected = ["hello", "world", "python"]
    test_expression(
        data=test_data,
        expected=expected,
        name="lower",
        namespace="str",
    )


def test_extract(test_expression):
    test_data = ["123-456", "789-012", "345-678"]
    regex = r"(\d)(\d*)"
    expected = ["123", "789", "345"]
    test_expression(
        data=test_data,
        expected=expected,
        name="extract",
        namespace="str",
        sql_name="regexp_extract",
        args=[regex],
    )


def test_substr(test_expression):
    test_data = ["daft", "query", "engine"]
    expected = [s[2:] for s in test_data]
    test_expression(
        data=test_data,
        expected=expected,
        name="substr",
        namespace="str",
        sql_name="substr",
        args=[2, None],
    )



from __future__ import annotations

import pyarrow as pa
import pytest

import daft
from daft import DataType, col, list_
from daft.daft import CountMode
from daft.sql.sql import SQLCatalog


def assert_eq(actual, expect):
    """Asserts two dataframes are equal for tests."""
    assert actual.collect().to_pydict() == expect.collect().to_pydict()


def test_list_constructor_empty():
    with pytest.raises(Exception, match="List constructor requires at least one item"):
        df = daft.from_pydict({"x": [1, 2, 3]})
        daft.sql("SELECT [ ] as list FROM df")
        df  # for ruff ignore unused


def test_list_constructor_different_lengths():
    with pytest.raises(Exception, match="Expected all columns to be of the same length"):
        df = daft.from_pydict({"x": [1, 2], "y": [3]})
        daft.sql("SELECT [ x, y ] FROM df")
        df  # for ruff ignore unused


def test_list_constructor_singleton():
    df = daft.from_pydict({"x": [1, 2, 3]})
    actual = daft.sql("SELECT [ x ] as list FROM df")
    expect = df.select(col("x").apply(lambda x: [x], DataType.list(DataType.int64())).alias("list"))
    assert_eq(actual, expect)


def test_list_constructor_homogeneous():
    df = daft.from_pydict({"x": [1, 2, 3]})
    actual = daft.sql("SELECT [ x * 1, x * 2, x * 3 ] FROM df")
    expect = df.select(col("x").apply(lambda x: [x * 1, x * 2, x * 3], DataType.list(DataType.int64())).alias("list"))
    assert_eq(actual, expect)


def test_list_constructor_heterogeneous():
    df = daft.from_pydict({"x": [1, 2, 3], "y": [True, True, False]})
    df = daft.sql("SELECT [ x, y ] AS heterogeneous FROM df").collect()
    assert df.to_pydict() == {"heterogeneous": [[1, 1], [2, 1], [3, 0]]}


def test_list_constructor_heterogeneous_with_cast():
    df = daft.from_pydict({"x": [1, 2, 3], "y": [True, True, False]})
    actual = daft.sql("SELECT [ CAST(x AS STRING), CAST(y AS STRING) ] FROM df")
    expect = df.select(list_(col("x").cast(DataType.string()), col("y").cast(DataType.string())))
    assert_eq(actual, expect)


def test_list_constructor_mixed_null_first():
    df = daft.from_pydict({"x": [1, 2, 3]})
    actual = daft.sql("SELECT [ NULL, x ] FROM df")
    expect = df.select(col("x").apply(lambda x: [None, x], DataType.list(DataType.int64())).alias("list"))
    assert_eq(actual, expect)


def test_list_constructor_mixed_null_mid():
    df = daft.from_pydict({"x": [1, 2, 3]})
    actual = daft.sql("SELECT [ x * -1, NULL, x ] FROM df")
    expect = df.select(col("x").apply(lambda x: [x * -1, None, x], DataType.list(DataType.int64())).alias("list"))
    assert_eq(actual, expect)


def test_list_constructor_mixed_null_last():
    df = daft.from_pydict({"x": [1, 2, 3]})
    actual = daft.sql("SELECT [ x, NULL ] FROM df")
    expect = df.select(col("x").apply(lambda x: [x, None], DataType.list(DataType.int64())).alias("list"))
    assert_eq(actual, expect)


def test_list_constructor_all_nulls():
    df = daft.from_pydict({"x": [1, 2, 3]})
    actual = daft.sql("SELECT [ NULL, NULL ] FROM df")
    expect = df.select(col("x").apply(lambda x: [None, None], DataType.list(DataType.null())).alias("list"))
    assert_eq(actual, expect)


def test_list_chunk():
    df = daft.from_pydict(
        {
            "col": pa.array([], type=pa.list_(pa.int64())),
            "fixed_col": pa.array([], type=pa.list_(pa.int64(), 2)),
        }
    )
    catalog = SQLCatalog({"test": df})
    expected = df.select(
        col("col").list.chunk(1).alias("col1"),
        col("col").list.chunk(2).alias("col2"),
        col("col").list.chunk(1000).alias("col3"),
        col("fixed_col").list.chunk(1).alias("fixed_col1"),
        col("fixed_col").list.chunk(2).alias("fixed_col2"),
        col("fixed_col").list.chunk(1000).alias("fixed_col3"),
    )

    actual = daft.sql(
        """
    SELECT
        list_chunk(col, 1) as col1,
        list_chunk(col, 2) as col2,
        list_chunk(col, 1000) as col3,
        list_chunk(fixed_col, 1) as fixed_col1,
        list_chunk(fixed_col, 2) as fixed_col2,
        list_chunk(fixed_col, 1000) as fixed_col3
    FROM test
    """,
        catalog=catalog,
    ).collect()
    assert actual.to_pydict() == expected.to_pydict()


def test_list_counts():
    df = daft.from_pydict({"col": [[1, 2, 3], [1, 2], [1, None, 4], []]})
    catalog = SQLCatalog({"test": df})
    expected = df.select(
        col("col").list.count().alias("count_valid"),
        col("col").list.count(CountMode.All).alias("count_all"),
        col("col").list.count(CountMode.Null).alias("count_null"),
    ).collect()
    actual = daft.sql(
        """
    SELECT
        list_count(col) as count_valid,
        list_count(col, 'all') as count_all,
        list_count(col, 'null') as count_null
    FROM test
    """,
        catalog=catalog,
    ).collect()
    assert actual.to_pydict() == expected.to_pydict()


def test_list_explode():
    df = daft.from_pydict({"col": [[1, 2, 3], [1, 2], [1, None, 4], []]})
    catalog = SQLCatalog({"test": df})
    expected = df.explode(col("col"))
    actual = daft.sql("SELECT unnest(col) as col FROM test", catalog=catalog).collect()
    assert actual.to_pydict() == expected.to_pydict()
    # test with alias
    actual = daft.sql("SELECT explode(col) as col FROM test", catalog=catalog).collect()
    assert actual.to_pydict() == expected.to_pydict()


def test_list_join():
    df = daft.from_pydict({"col": [None, [], ["a"], [None], ["a", "a"], ["a", None], ["a", None, "a"]]})
    catalog = SQLCatalog({"test": df})
    expected = df.select(col("col").list.join(","))
    actual = daft.sql("SELECT list_join(col, ',') FROM test", catalog=catalog).collect()
    assert actual.to_pydict() == expected.to_pydict()
    # make sure it works with the `array_to_string` function too
    actual = daft.sql("SELECT array_to_string(col, ',') FROM test", catalog=catalog).collect()
    assert actual.to_pydict() == expected.to_pydict()


def test_various_list_ops():
    df = daft.from_pydict({"col": [[1, 2, 3], [1, 2], [1, None, 4], []]})
    catalog = SQLCatalog({"test": df})
    expected = df.select(
        col("col").list.min().alias("min"),
        col("col").list.max().alias("max"),
        col("col").list.mean().alias("mean"),
        col("col").list.sum().alias("sum"),
        col("col").list.sort().alias("sort"),
        col("col").list.sort(True).alias("sort_desc"),
        col("col").list.sort(False).alias("sort_asc"),
        col("col").list.slice(1, 2).alias("slice"),
    ).collect()
    actual = daft.sql(
        """
    SELECT
        list_min(col) as min,
        list_max(col) as max,
        list_mean(col) as mean,
        list_sum(col) as sum,
        list_sort(col) as sort,
        list_sort(col, desc:=true) as sort_desc,
        list_sort(col, desc:=false) as sort_asc,
        list_slice(col, 1, 2) as slice
    FROM test
    """,
        catalog=catalog,
    ).collect()
    assert actual.to_pydict() == expected.to_pydict()



from __future__ import annotations

import os
import shutil
import tempfile

import pytest

import daft
from daft import DataFrame
from tests.conftest import get_tests_daft_runner_name


@pytest.fixture(scope="module", params=[(1, 64), (8, 8), (64, 1)], ids=["1x64mib", "8x8mib", "64x1mib"])
def gen_simple_csvs(request) -> str:
    """Creates some CSVs in a directory. Returns the name of the directory."""
    num_files, mibs_per_file = request.param

    _8bytes = b"aaa,bbb\n"
    _1kib = _8bytes * 128
    _1mib = _1kib * 1024

    with tempfile.TemporaryDirectory() as tmpdirname:
        # Make one CSV file of the correct size.
        with open(os.path.join(tmpdirname, "file.csv"), "wb") as f:
            f.write(b"A,B\n")
            for i in range(mibs_per_file):
                f.write(_1mib)

        # Copy it to get the remaining number of desired files.
        for i in range(1, num_files):
            shutil.copyfile(
                src=os.path.join(tmpdirname, "file.csv"),
                dst=os.path.join(tmpdirname, f"file{i}.csv"),
            )

        yield tmpdirname, num_files * mibs_per_file * 1024 * 128


@pytest.mark.skipif(get_tests_daft_runner_name() != "native", reason="requires local runner")
@pytest.mark.benchmark(group="file_read")
def test_csv_read(gen_simple_csvs, benchmark):
    csv_dir, num_rows = gen_simple_csvs

    def bench() -> DataFrame:
        df = daft.read_csv(csv_dir)
        return df.collect()

    df = benchmark(bench)

    assert len(df) == num_rows


@pytest.mark.benchmark(group="file_read")
@pytest.mark.parametrize("prune", [True, False])
def test_s3_parquet_read_1x64mb(benchmark, prune):
    parquet_glob = "s3://daft-public-data/test_fixtures/parquet/95c7fba0-265d-440b-88cb-2897047fc5f9-0.parquet"
    expected_rows = 1500000

    def bench() -> DataFrame:
        df = daft.read_parquet(parquet_glob)
        if prune:
            df = df.select(df["O_SHIPPRIORITY"])  # rightmost int64 column
        return df.collect()

    df = benchmark(bench)
    assert len(df.to_pandas()) == expected_rows


@pytest.mark.benchmark(group="file_read")
@pytest.mark.parametrize("prune", [True, False])
def test_s3_parquet_read_32x2mb(benchmark, prune):
    parquet_glob = "s3://daft-public-data/test_fixtures/parquet_small/*"
    expected_rows = 2000000

    def bench() -> DataFrame:
        df = daft.read_parquet(parquet_glob)
        if prune:
            df = df.select(df["P_SIZE"])  # rightmost int64 column
        return df.collect()

    df = benchmark(bench)
    assert len(df.to_pandas()) == expected_rows



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import boto3

    from daft.daft import S3Config


def boto3_client_from_s3_config(service: str, s3_config: S3Config) -> boto3.client:
    import boto3

    return boto3.client(
        service,
        region_name=s3_config.region_name,
        use_ssl=s3_config.use_ssl,
        verify=s3_config.verify_ssl,
        endpoint_url=s3_config.endpoint_url,
        aws_access_key_id=s3_config.key_id,
        aws_secret_access_key=s3_config.access_key,
        aws_session_token=s3_config.session_token,
    )



from __future__ import annotations

import copy
from datetime import date, datetime, time, timedelta, timezone

import pytest
import pytz

import daft
from daft.datatype import DataType, TimeUnit
from daft.expressions import col, lit
from daft.expressions.testing import expr_structurally_equal
from daft.recordbatch import MicroPartition
from daft.series import Series


@pytest.mark.parametrize(
    "data, expected_dtype",
    [
        (1, DataType.int32()),
        (2**32, DataType.int64()),
        (1 << 63, DataType.uint64()),
        (1.2, DataType.float64()),
        ("a", DataType.string()),
        (b"a", DataType.binary()),
        (True, DataType.bool()),
        (None, DataType.null()),
        (Series.from_pylist([1, 2, 3]), DataType.list(DataType.int64())),
        (date(2023, 1, 1), DataType.date()),
        (time(1, 2, 3, 4), DataType.time(timeunit=TimeUnit.from_str("us"))),
        (datetime(2023, 1, 1), DataType.timestamp(timeunit=TimeUnit.from_str("us"))),
        (
            datetime(2022, 1, 1, tzinfo=pytz.utc),
            DataType.timestamp(timeunit=TimeUnit.from_str("us"), timezone="UTC"),
        ),
    ],
)
def test_make_lit(data, expected_dtype) -> None:
    literal = lit(data)
    assert literal.name() == "literal"
    empty_table = MicroPartition.empty()
    lit_table = empty_table.eval_expression_list([literal])
    series = lit_table.get_column_by_name("literal")
    assert series.datatype() == expected_dtype
    repr_out = repr(literal)

    assert repr_out.startswith("lit(")
    assert repr_out.endswith(")")
    copied = copy.deepcopy(literal)
    assert repr_out == repr(copied)


import operator as ops

OPS = [
    (ops.add, "+"),
    (ops.sub, "-"),
    (ops.mul, "*"),
    (ops.truediv, "/"),
    (ops.mod, "%"),
    (ops.lt, "<"),
    (ops.le, "<="),
    (ops.eq, "=="),
    (ops.ne, "!="),
    (ops.ge, ">="),
    (ops.gt, ">"),
]


@pytest.mark.parametrize("op, symbol", OPS)
def test_repr_binary_operators(op, symbol) -> None:
    a = col("a")
    b = col("b")
    y = op(a, b)
    output = repr(y)
    tokens = output.split(" ")
    assert len(tokens) == 3
    assert tokens[0] == "col(a)"
    assert tokens[1] == symbol
    assert tokens[2] == "col(b)"
    copied = copy.deepcopy(y)
    assert output == repr(copied)


def test_repr_functions_abs() -> None:
    a = col("a")
    y = abs(a)
    repr_out = repr(y)
    assert repr_out == "abs(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_ceil() -> None:
    a = col("a")
    y = a.ceil()
    repr_out = repr(y)
    assert repr_out == "ceil(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_floor() -> None:
    a = col("a")
    y = a.floor()
    repr_out = repr(y)
    assert repr_out == "floor(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_sign() -> None:
    a = col("a")
    y = a.sign()
    repr_out = repr(y)
    assert repr_out == "sign(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_signum() -> None:
    a = col("a")
    y = a.signum()
    repr_out = repr(y)
    assert repr_out == "sign(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_negate() -> None:
    a = col("a")
    y = a.negate()
    repr_out = repr(y)
    assert repr_out == "negative(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_negative() -> None:
    a = col("a")
    y = a.negative()
    repr_out = repr(y)
    assert repr_out == "negative(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_round() -> None:
    a = col("a")
    y = a.round()
    repr_out = repr(y)
    assert repr_out == "round(col(a), lit(0))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_clip() -> None:
    a = col("a")
    b = col("b")
    c = col("c")
    y = a.clip(b, c)
    repr_out = repr(y)
    assert repr_out == "clip(col(a), col(b), col(c))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_log2() -> None:
    a = col("a")
    y = a.log2()
    repr_out = repr(y)
    assert repr_out == "log2(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_log10() -> None:
    a = col("a")
    y = a.log10()
    repr_out = repr(y)
    assert repr_out == "log10(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_log() -> None:
    a = col("a")
    y = a.log(2)
    repr_out = repr(y)
    assert repr_out == "log(col(a), lit(2))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_ln() -> None:
    a = col("a")
    y = a.ln()
    repr_out = repr(y)
    assert repr_out == "ln(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_log1p() -> None:
    a = col("a")
    y = a.log1p()
    repr_out = repr(y)
    assert repr_out == "log1p(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_shift_left() -> None:
    a = col("a")
    b = col("b")
    y = a << (b)
    repr_out = repr(y)
    assert repr_out == "col(a) << col(b)"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_shift_right() -> None:
    a = col("a")
    b = col("b")
    y = a >> (b)
    repr_out = repr(y)
    assert repr_out == "col(a) >> col(b)"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


@pytest.mark.parametrize(
    "fun",
    [
        "sin",
        "cos",
        "tan",
        "csc",
        "sec",
        "cot",
        "sinh",
        "cosh",
        "tanh",
        "arcsin",
        "arccos",
        "arctan",
        "radians",
        "degrees",
    ],
)
def test_repr_functions_trigonometry(fun: str) -> None:
    a = col("a")
    y = getattr(a, fun)()
    repr_out = repr(y)
    assert repr_out == f"{fun}(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_atan2() -> None:
    a = col("a")
    b = col("b")
    y = a.arctan2(b)
    repr_out = repr(y)
    assert repr_out == "atan2(col(a), col(b))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_arctanh() -> None:
    a = col("a")
    y = a.arctanh()
    repr_out = repr(y)
    assert repr_out == "arctanh(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_arccosh() -> None:
    a = col("a")
    y = a.arccosh()
    repr_out = repr(y)
    assert repr_out == "arccosh(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_arcsinh() -> None:
    a = col("a")
    y = a.arcsinh()
    repr_out = repr(y)
    assert repr_out == "arcsinh(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_unix_date() -> None:
    a = col("a")
    y = a.dt.unix_date()
    repr_out = repr(y)
    assert repr_out == "unix_date(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_day() -> None:
    a = col("a")
    y = a.dt.day()
    repr_out = repr(y)
    assert repr_out == "day(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_month() -> None:
    a = col("a")
    y = a.dt.month()
    repr_out = repr(y)
    assert repr_out == "month(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_quarter() -> None:
    a = col("a")
    y = a.dt.quarter()
    repr_out = repr(y)
    assert repr_out == "quarter(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_year() -> None:
    a = col("a")
    y = a.dt.year()
    repr_out = repr(y)
    assert repr_out == "year(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_day_of_week() -> None:
    a = col("a")
    y = a.dt.day_of_week()
    repr_out = repr(y)
    assert repr_out == "day_of_week(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_day_of_month() -> None:
    a = col("a")
    y = a.dt.day_of_month()
    repr_out = repr(y)
    assert repr_out == "day_of_month(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_day_of_year() -> None:
    a = col("a")
    y = a.dt.day_of_year()
    repr_out = repr(y)
    assert repr_out == "day_of_year(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_week_of_year() -> None:
    a = col("a")
    y = a.dt.week_of_year()
    repr_out = repr(y)
    assert repr_out == "week_of_year(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_exp() -> None:
    a = col("a")
    y = a.exp()
    repr_out = repr(y)
    assert repr_out == "exp(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_expm1() -> None:
    a = col("a")
    y = a.expm1()
    repr_out = repr(y)
    assert repr_out == "expm1(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_sqrt() -> None:
    a = col("a")
    y = a.sqrt()
    repr_out = repr(y)
    assert repr_out == "sqrt(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_cbrt() -> None:
    a = col("a")
    y = a.cbrt()
    repr_out = repr(y)
    assert repr_out == "cbrt(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_hash() -> None:
    a = col("a")
    y = a.hash()
    repr_out = repr(y)
    assert repr_out == "hash(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_hash_2() -> None:
    a = col("a")
    y = a.hash(lit(1))
    repr_out = repr(y)
    assert repr_out == "hash(col(a), lit(1))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_minhash() -> None:
    a = col("a")
    y = a.minhash(1, 2)
    repr_out = repr(y)
    assert repr_out == "minhash(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_minhash_2() -> None:
    a = col("a")
    y = a.minhash(1, 2, 3)
    repr_out = repr(y)
    assert repr_out == "minhash(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_tokenize_encode() -> None:
    a = col("a")
    y = a.str.tokenize_encode("cl100k_base")
    repr_out = repr(y)
    assert repr_out == "tokenize_encode(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_repr_functions_tokenize_decode() -> None:
    a = col("a")
    y = a.str.tokenize_decode("cl100k_base")
    repr_out = repr(y)
    assert repr_out == "tokenize_decode(col(a))"
    copied = copy.deepcopy(y)
    assert repr_out == repr(copied)


def test_expr_structurally_equal() -> None:
    e1 = (col("a").max() == col("b").alias("moo") - 3).is_null()
    e2 = (col("a").max() == col("b").alias("moo") - 3).is_null()
    e3 = (col("a").max() == col("b").alias("moo") - 4).is_null()
    assert expr_structurally_equal(e1, e2)
    assert not expr_structurally_equal(e2, e3)


def test_str_concat_delegation() -> None:
    a = col("a")
    b = "foo"
    c = a.str.concat(b)
    expected = a + lit(b)
    assert expr_structurally_equal(c, expected)
    output = repr(c)
    assert output == 'col(a) + lit("foo")'


def test_float_is_nan() -> None:
    a = col("a")
    c = a.float.is_nan()
    output = repr(c)
    assert output == "is_nan(col(a))"


def test_float_is_inf() -> None:
    a = col("a")
    c = a.float.is_inf()
    output = repr(c)
    assert output == "is_inf(col(a))"


def test_float_not_nan() -> None:
    a = col("a")
    c = a.float.not_nan()
    output = repr(c)
    assert output == "not_nan(col(a))"


def test_date_lit_post_epoch() -> None:
    d = lit(date(2022, 1, 1))
    output = repr(d)
    assert output == "lit(2022-01-01)"


def test_date_lit_pre_epoch() -> None:
    d = lit(date(1950, 1, 1))
    output = repr(d)
    assert output == "lit(1950-01-01)"


@pytest.mark.parametrize(
    "input, expected",
    [
        (
            datetime(2022, 1, 1, 12, 30, 59, 0),
            "lit(2022-01-01 12:30:59)",
        ),
        (
            datetime(2022, 1, 1, 12, 30, 59, 123456),
            "lit(2022-01-01 12:30:59.123456)",
        ),
    ],
)
def test_datetime_lit_post_epoch(input, expected) -> None:
    d = lit(input)
    output = repr(d)
    assert output == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        (
            pytz.timezone("US/Eastern").localize(datetime(2022, 1, 1, 12, 30, 59, 0)),
            "lit(2022-01-01 12:30:59 EST)",
        ),
        (
            datetime(2022, 1, 1, 12, 30, 59, 123456, tzinfo=timezone(timedelta(hours=9))),
            "lit(2022-01-01 12:30:59.123456 +09:00)",
        ),
    ],
)
def test_datetime_tz_lit_post_epoch(input, expected) -> None:
    d = lit(input)
    output = repr(d)
    assert output == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        (
            datetime(1950, 1, 1, 12, 30, 59, 0),
            "lit(1950-01-01 12:30:59)",
        ),
        (
            datetime(1950, 1, 1, 12, 30, 59, 123456),
            "lit(1950-01-01 12:30:59.123456)",
        ),
    ],
)
def test_datetime_lit_pre_epoch(input, expected) -> None:
    d = lit(input)
    output = repr(d)
    assert output == expected


@pytest.mark.parametrize(
    "timeunit, expected",
    [
        (
            "s",
            "1970-01-01 00:00:01",
        ),
        (
            "ms",
            "1970-01-01 00:00:00.001",
        ),
        (
            "us",
            "1970-01-01 00:00:00.000001",
        ),
        (
            "ns",
            "1970-01-01 00:00:00.000000001",
        ),
    ],
)
def test_datetime_lit_different_timeunits(timeunit, expected) -> None:
    import pyarrow as pa

    pa_array = pa.array([1], type=pa.timestamp(timeunit))
    pa_table = pa.table({"dt": pa_array})
    mp = MicroPartition.from_arrow(pa_table)

    series = mp.get_column_by_name("dt")
    output = repr(series)

    # Series repr is very long, so we just check the last line which contains the timestamp
    # Also remove the │ │, which is what the [1:-1] and strip is doing
    timestamp_repr = output.split("\n")[-3][1:-1].strip()
    assert timestamp_repr == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        (
            timedelta(days=1),
            "lit(1d)",
        ),
        (
            timedelta(days=1, hours=12, minutes=30, seconds=59),
            "lit(1d 12h 30m 59s)",
        ),
        (
            timedelta(days=1, hours=12, minutes=30, seconds=59, microseconds=123456),
            "lit(1d 12h 30m 59s 123456µs)",
        ),
        (
            timedelta(weeks=1, days=1, hours=12, minutes=30, seconds=59, microseconds=123456),
            "lit(8d 12h 30m 59s 123456µs)",
        ),
    ],
)
def test_duration_lit(input, expected) -> None:
    d = lit(input)
    output = repr(d)
    assert output == expected


def test_repr_series_lit() -> None:
    s = lit(Series.from_pylist([1, 2, 3]))
    output = repr(s)
    assert output == "lit([[1, 2, 3]])"


def test_list_value_counts():
    # Create a MicroPartition with a list column
    mp = MicroPartition.from_pydict(
        {"list_col": [["a", "b", "a", "c"], ["b", "b", "c"], ["a", "a", "a"], [], ["d", None, "d"]]}
    )

    # Apply list_value_counts operation
    result = mp.eval_expression_list([col("list_col").list.value_counts().alias("value_counts")])
    value_counts = result.to_pydict()["value_counts"]

    # Expected output
    expected = [[("a", 2), ("b", 1), ("c", 1)], [("b", 2), ("c", 1)], [("a", 3)], [], [("d", 2)]]

    # Check the result
    assert value_counts == expected

    # Test with empty input (no proper type -> should raise error)
    empty_mp = MicroPartition.from_pydict({"list_col": []})
    with pytest.raises(ValueError):
        empty_mp.eval_expression_list([col("list_col").list.value_counts().alias("value_counts")])

    # Test with empty input (no proper type -> should raise error)
    none_mp = MicroPartition.from_pydict({"list_col": [None, None, None]})
    with pytest.raises(ValueError):
        none_mp.eval_expression_list([col("list_col").list.value_counts().alias("value_counts")])


def test_list_value_counts_nested():
    # Create a MicroPartition with a nested list column
    mp = MicroPartition.from_pydict(
        {
            "nested_list_col": [
                [[1, 2], [3, 4]],
                [[1, 2], [5, 6]],
                [[3, 4], [1, 2]],
                [],
                None,
                [[1, 2], [1, 2]],
            ]
        }
    )

    # Apply list_value_counts operation and expect an exception
    with pytest.raises(daft.exceptions.DaftCoreException) as exc_info:
        mp.eval_expression_list([col("nested_list_col").list.value_counts().alias("value_counts")])

    # Check the exception message
    assert (
        'DaftError::ArrowError Invalid argument error: The data type type LargeList(Field { name: "item", data_type: Int64, is_nullable: true, metadata: {} }) has no natural order'
        in str(exc_info.value)
    )


def test_list_value_counts_fixed_size():
    # Create data with lists of fixed size
    data = {
        "fixed_list": [
            [1, 2, 3],
            [4, 3, 4],
            [4, 5, 6],
            [1, 2, 3],
            [7, 8, 9],
            None,
        ]
    }

    # Create DataFrame and cast the column to fixed size list
    df = daft.from_pydict(data).with_column(
        "fixed_list", daft.col("fixed_list").cast(DataType.fixed_size_list(DataType.int64(), 3))
    )

    df = df.with_column("fixed_list", col("fixed_list").cast(DataType.fixed_size_list(DataType.int64(), 3)))

    # Get value counts
    result = df.with_column("value_counts", col("fixed_list").list.value_counts())

    # Verify the value counts
    result_dict = result.to_pydict()
    assert result_dict["value_counts"] == [
        [(1, 1), (2, 1), (3, 1)],
        [(4, 2), (3, 1)],
        [(4, 1), (5, 1), (6, 1)],
        [(1, 1), (2, 1), (3, 1)],
        [(7, 1), (8, 1), (9, 1)],
        [],
    ]


def test_list_value_counts_degenerate():
    import pyarrow as pa

    # Create a MicroPartition with an empty list column of specified type
    empty_mp = MicroPartition.from_pydict({"empty_list_col": pa.array([], type=pa.list_(pa.string()))})

    # Apply list_value_counts operation
    result = empty_mp.eval_expression_list([col("empty_list_col").list.value_counts().alias("value_counts")])

    # Check the result
    assert result.to_pydict() == {"value_counts": []}

    # Test with null values
    null_mp = MicroPartition.from_pydict({"null_list_col": pa.array([None, None], type=pa.list_(pa.string()))})

    result_null = null_mp.eval_expression_list([col("null_list_col").list.value_counts().alias("value_counts")])

    # Check the result for null values
    assert result_null.to_pydict() == {"value_counts": [[], []]}


@pytest.mark.parametrize(
    # we don't need to test all types, just a few as a sanity check. The rest are tested in the sql tests
    "sql, actual",
    [
        ("int32", DataType.int32()),
        ("int64", DataType.int64()),
        ("float32", DataType.float32()),
        ("real", DataType.float32()),
        ("float(5)", DataType.float32()),
        ("float64", DataType.float64()),
        ("float", DataType.float64()),
        ("double", DataType.float64()),
        ("double 10", DataType.float64()),
        ("float(30)", DataType.float64()),
        ("text", DataType.string()),
        ("string", DataType.string()),
        ("varchar", DataType.string()),
    ],
)
def test_cast_sql_string(sql, actual):
    expr = col("a").cast(sql)
    actual = col("a").cast(actual)
    df = daft.from_pydict({"a": [1, 2, 3]}).select(expr)
    actual_df = daft.from_pydict({"a": [1, 2, 3]}).select(actual)
    assert df.schema() == actual_df.schema()


@pytest.mark.parametrize(
    # Some sql expressions were failing as documented previously
    "sql",
    ["char", "character", "char varying", "character varying"],
)
def test_cast_sql_string_failing(sql):
    with pytest.raises(Exception):
        expr = col("a").cast(sql)
        daft.from_pydict({"a": [1, 2, 3]}).select(expr)


def test_drop_nan():
    dd = {"n": [1.0, 2.0, None]}
    df = daft.from_pydict(dd)
    assert df.drop_nan().to_pydict() == {"n": [1.0, 2.0]}


def test_drop_nan_no_nans():
    dd = {"n": [1, 1, 2]}
    df = daft.from_pydict(dd)
    assert df.drop_nan().to_pydict() == dd



from __future__ import annotations

from pathlib import Path

import pytest
from fsspec.implementations.http import HTTPFileSystem

from daft.daft import io_glob
from tests.integration.io.conftest import mount_data_nginx


def compare_http_result(daft_ls_result: list, fsspec_result: list):
    daft_files = [(f["path"], f["type"].lower(), f["size"]) for f in daft_ls_result]
    httpfs_files = [(f["name"], f["type"], f["size"]) for f in fsspec_result]

    # io_glob doesn't return directory entries
    httpfs_files = [(p, t, s) for p, t, s in httpfs_files if t == "file" and not p.endswith("/")]

    assert len(daft_files) == len(httpfs_files)
    assert sorted(daft_files) == sorted(httpfs_files)


@pytest.fixture(scope="module")
def nginx_http_url(nginx_config, tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("test-list-http")
    data_path = Path(tmpdir)
    (Path(data_path) / "file.txt").touch()
    (Path(data_path) / "test_ls").mkdir()
    (Path(data_path) / "test_ls" / "file.txt").touch()
    (Path(data_path) / "test_ls" / "paginated-10-files").mkdir()
    for i in range(10):
        (Path(data_path) / "test_ls" / "paginated-10-files" / f"file.{i}.txt").touch()

    with mount_data_nginx(nginx_config, data_path):
        yield nginx_config[0]


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path",
    [
        "",
        "/",
        "/test_ls",
        "/test_ls/",
        "/test_ls//",
        "/test_ls/paginated-10-files/",
    ],
)
def test_http_flat_directory_listing(path, nginx_http_url):
    http_path = f"{nginx_http_url}{path}"
    fs = HTTPFileSystem()
    fsspec_result = fs.ls(http_path, detail=True)
    daft_ls_result = io_glob(http_path)
    compare_http_result(daft_ls_result, fsspec_result)


@pytest.mark.integration()
def test_http_single_file_listing(nginx_http_url):
    path = f"{nginx_http_url}/test_ls/file.txt"
    daft_ls_result = io_glob(path)

    # NOTE: FSSpec will return size 0 list for this case, but we want to return 1 element to be
    # consistent with behavior of our other file listing utilities
    # fs = HTTPFileSystem()
    # fsspec_result = fs.ls(path, detail=True)

    assert len(daft_ls_result) == 1
    assert daft_ls_result[0] == {"path": path, "size": 0, "type": "File"}


@pytest.mark.integration()
def test_http_notfound(nginx_http_url):
    path = f"{nginx_http_url}/test_ls/MISSING"
    fs = HTTPFileSystem()
    with pytest.raises(FileNotFoundError, match=path):
        fs.ls(path, detail=True)

    with pytest.raises(FileNotFoundError, match=path):
        io_glob(path)


@pytest.mark.integration()
def test_http_flat_directory_listing_recursive(nginx_http_url):
    http_path = f"{nginx_http_url}/**"
    fs = HTTPFileSystem()
    fsspec_result = list(fs.glob(http_path, detail=True).values())
    daft_ls_result = io_glob(http_path)
    compare_http_result(daft_ls_result, fsspec_result)


@pytest.mark.integration()
def test_http_listing_absolute_urls(nginx_config, tmpdir):
    nginx_http_url, _ = nginx_config

    tmpdir = Path(tmpdir)
    test_manifest_file = tmpdir / "index.html"
    test_manifest_file.write_text(
        f"""
        <a href="{nginx_http_url}/other.html">this is an absolute path to a file</a>
        <a href="{nginx_http_url}/dir/">this is an absolute path to a dir</a>
    """
    )

    with mount_data_nginx(nginx_config, tmpdir):
        http_path = f"{nginx_http_url}/"
        daft_ls_result = io_glob(http_path)

        # NOTE: Cannot use fsspec here because they do not correctly find the links
        # fsspec_result = fs.ls(http_path, detail=True)
        # compare_http_result(daft_ls_result, fsspec_result)

        assert daft_ls_result == [
            {"type": "File", "path": f"{nginx_http_url}/other.html", "size": None},
        ]


@pytest.mark.integration()
def test_http_listing_absolute_base_urls(nginx_config, tmpdir):
    nginx_http_url, _ = nginx_config

    tmpdir = Path(tmpdir)
    test_manifest_file = tmpdir / "index.html"
    test_manifest_file.write_text(
        """
        <a href="/other.html">this is an absolute base path to a file</a>
        <a href="/dir/">this is an absolute base path to a dir</a>
    """
    )

    with mount_data_nginx(nginx_config, tmpdir):
        http_path = f"{nginx_http_url}/"
        daft_ls_result = io_glob(http_path)

        # NOTE: Cannot use fsspec here because they do not correctly find the links
        # fsspec_result = fs.ls(http_path, detail=True)
        # compare_http_result(daft_ls_result, fsspec_result)

        assert daft_ls_result == [
            {"type": "File", "path": f"{nginx_http_url}/other.html", "size": None},
        ]



from __future__ import annotations

import daft
from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_to_struct():
    df = MicroPartition.from_pydict(
        {
            "a": [1, 2, 3, 4, None, 6, None],
            "b": ["a", "b", "c", "", "e", None, None],
        }
    )
    res = df.eval_expression_list([daft.struct(col("a"), col("b"))]).to_pydict()
    assert res["struct"] == [
        {"a": 1, "b": "a"},
        {"a": 2, "b": "b"},
        {"a": 3, "b": "c"},
        {"a": 4, "b": ""},
        {"a": None, "b": "e"},
        {"a": 6, "b": None},
        {"a": None, "b": None},
    ]


def test_to_struct_subexpr():
    df = MicroPartition.from_pydict(
        {
            "a": [1, 2, 3, 4, None, 6, None],
            "b": ["a", "b", "c", "", "e", None, None],
        }
    )
    res = df.eval_expression_list([daft.struct(col("a") * 2, col("b").str.upper())]).to_pydict()
    assert res["struct"] == [
        {"a": 2, "b": "A"},
        {"a": 4, "b": "B"},
        {"a": 6, "b": "C"},
        {"a": 8, "b": ""},
        {"a": None, "b": "E"},
        {"a": 12, "b": None},
        {"a": None, "b": None},
    ]


def test_to_struct_strs():
    df = MicroPartition.from_pydict(
        {
            "a": [1, 2, 3, 4, None, 6, None],
            "b": ["a", "b", "c", "", "e", None, None],
        }
    )
    res = df.eval_expression_list([daft.struct("a", "b")]).to_pydict()
    assert res["struct"] == [
        {"a": 1, "b": "a"},
        {"a": 2, "b": "b"},
        {"a": 3, "b": "c"},
        {"a": 4, "b": ""},
        {"a": None, "b": "e"},
        {"a": 6, "b": None},
        {"a": None, "b": None},
    ]



from __future__ import annotations

import pytest

from daft.recordbatch import MicroPartition


def test_table_concat() -> None:
    objs1 = [None, object(), object()]
    objs2 = [object(), None, object()]
    tables = [
        MicroPartition.from_pydict({"x": [1, 2, 3], "y": ["a", "b", "c"], "z": objs1}),
        MicroPartition.from_pydict({"x": [4, 5, 6], "y": ["d", "e", "f"], "z": objs2}),
    ]

    result = MicroPartition.concat(tables)
    assert result.to_pydict() == {"x": [1, 2, 3, 4, 5, 6], "y": ["a", "b", "c", "d", "e", "f"], "z": objs1 + objs2}

    tables = [
        MicroPartition.from_pydict({"x": [], "y": []}),
        MicroPartition.from_pydict({"x": [], "y": []}),
    ]

    result = MicroPartition.concat(tables)
    assert result.to_pydict() == {"x": [], "y": []}


def test_table_concat_bad_input() -> None:
    mix_types_table = [MicroPartition.from_pydict({"x": [1, 2, 3]}), []]
    with pytest.raises(TypeError, match=f"Expected a {MicroPartition.__name__} for concat"):
        MicroPartition.concat(mix_types_table)

    with pytest.raises(ValueError, match=f"Need at least 1 {MicroPartition.__name__}"):
        MicroPartition.concat([])


def test_table_concat_schema_mismatch() -> None:
    mix_types_table = [
        MicroPartition.from_pydict({"x": [1, 2, 3]}),
        MicroPartition.from_pydict({"y": [1, 2, 3]}),
    ]

    with pytest.raises(ValueError, match=f"{MicroPartition.__name__} concat requires all schemas to match"):
        MicroPartition.concat(mix_types_table)

    mix_types_table = [
        MicroPartition.from_pydict({"x": [1, 2, 3]}),
        MicroPartition.from_pydict({"x": [1.0, 2.0, 3.0]}),
    ]

    with pytest.raises(ValueError, match=f"{MicroPartition.__name__} concat requires all schemas to match"):
        MicroPartition.concat(mix_types_table)

    mix_types_table = [
        MicroPartition.from_pydict({"x": [1, 2, 3]}),
        MicroPartition.from_pydict({"x": [object(), object(), object()]}),
    ]

    with pytest.raises(ValueError, match=f"{MicroPartition.__name__} concat requires all schemas to match"):
        MicroPartition.concat(mix_types_table)

    mix_types_table = [
        MicroPartition.from_pydict({"x": [1, 2, 3]}),
        MicroPartition.from_pydict({"x": [1, 2, 3], "y": [2, 3, 4]}),
    ]

    with pytest.raises(ValueError, match=f"{MicroPartition.__name__} concat requires all schemas to match"):
        MicroPartition.concat(mix_types_table)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import boto3
import pytest
from moto import mock_aws

from daft import Catalog, Identifier
from daft.logical.schema import DataType as dt
from daft.logical.schema import Field, Schema

if TYPE_CHECKING:
    from mypy_boto3_s3tables import S3TablesClient
else:
    S3TablesClient = object


@pytest.mark.skip("skipped for integration testing")
def test_s3tables_iceberg_rest():
    import daft

    table_bucket_arn = "..."
    catalog = Catalog.from_s3tables(table_bucket_arn)
    print(catalog.list_tables("demo"))
    catalog.read_table("demo.points").show()
    catalog.write_table(
        "demo.points",
        daft.from_pydict(
            {
                "x": [True],
                "y": [4],
                "z": ["d"],
            }
        ),
    )
    catalog.read_table("demo.points").show()


@pytest.fixture(scope="function")
def aws_credentials():
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-west-2"


@pytest.fixture(scope="function")
def client(aws_credentials):
    with mock_aws():
        yield boto3.client("s3tables", region_name="us-west-2")


@pytest.fixture(scope="function")
def catalog(client: S3TablesClient):
    res = client.create_table_bucket(name="test_bucket")
    arn = res["arn"]
    yield Catalog.from_s3tables(arn, client=client)


def schema(fields: dict[str, dt]) -> Schema:
    return Schema._from_fields([Field.create(k, v) for k, v in fields.items()])


def test_create_namespace(catalog: Catalog):
    namespace = "ns"
    catalog.create_namespace(namespace)
    namespaces = catalog.list_namespaces()
    assert Identifier("ns") in namespaces


def test_create_table(catalog: Catalog):
    ns = "create_table_ns"
    table_name = f"{ns}.T"
    catalog.create_namespace(ns)
    table = catalog.create_table(
        table_name,
        schema(
            {
                "x": dt.bool(),
                "y": dt.int64(),
                "z": dt.string(),
            }
        ),
    )
    tables = catalog.list_tables()
    assert len(tables) == 1
    assert Identifier.from_str(table_name) in tables
    assert table.name == "T"


def test_drop(catalog: Catalog):
    ns = "ns_to_drop"
    table_name = f"{ns}.table_to_drop"
    catalog.create_namespace(ns)
    catalog.create_table(table_name, schema({"a": dt.bool()}))
    #
    assert Identifier(ns) in catalog.list_namespaces()
    assert Identifier.from_str(table_name) in catalog.list_tables(ns)
    #
    catalog.drop_table(table_name)
    assert Identifier.from_str(table_name) not in catalog.list_tables(ns)
    #
    catalog.drop_namespace(ns)
    assert Identifier(ns) not in catalog.list_namespaces()


def test_get_table(catalog: Catalog):
    ns = "ns_to_get"
    table_name = f"{ns}.table_to_get"
    catalog.create_namespace(ns)
    catalog.create_table(table_name, schema({"a": dt.bool()}))
    # verify
    res = catalog.get_table(table_name)
    assert res.name == "table_to_get"


def test_list_namespaces(catalog: Catalog):
    namespaces = [
        Identifier("ns1_1"),
        Identifier("ns1_2"),
        Identifier("ns2_1"),
    ]
    for ns in namespaces:
        catalog.create_namespace(ns)
    #
    # list one
    res = catalog.list_namespaces("ns1_1")
    assert len(res) == 1
    assert Identifier("ns1_1") == res[0]
    #
    # list some
    res = catalog.list_namespaces("ns1_")
    assert len(res) == 2
    assert Identifier("ns1_1") in res
    assert Identifier("ns1_2") in res
    #
    # list all
    res = catalog.list_namespaces()
    for ns in namespaces:
        assert ns in res


def test_list_tables(catalog: Catalog):
    ns1 = "ns1"
    ns2 = "ns2"
    tables = [
        f"{ns1}.tbl1_1",
        f"{ns1}.tbl1_2",
        f"{ns1}.tbl2_1",
        f"{ns2}.tbl1_1",
        f"{ns2}.tbl1_2",
        f"{ns2}.tbl2_1",
    ]
    catalog.create_namespace(ns1)
    catalog.create_namespace(ns2)
    for table in tables:
        catalog.create_table(table, schema({}))
    #
    # list one
    res = catalog.list_tables(f"{ns1}.tbl1_1")
    assert len(res) == 1
    assert Identifier(ns1, "tbl1_1") == res[0]
    #
    # list some
    res = catalog.list_tables(ns1)
    assert len(res) == 3
    assert Identifier(ns1, "tbl1_1") in res
    assert Identifier(ns1, "tbl1_2") in res
    assert Identifier(ns1, "tbl2_1") in res
    #
    # list all
    res = catalog.list_tables()
    assert len(res) == len(tables)
    for table in tables:
        assert Identifier.from_str(table) in res


def test_existence_checks(catalog: Catalog):
    ns1 = "ns1"
    catalog.create_namespace(ns1)

    assert catalog.has_namespace(ns1)
    assert not catalog.has_namespace("does_not_exist")


@pytest.mark.skip("S3 Tables read tests are done via integration tests.")
def test_read_table():
    pass  # https://github.com/Eventual-Inc/Daft/issues/3925


@pytest.mark.skip("S3 Tables read tests are done via integration tests.")
def test_write_table():
    pass  # https://github.com/Eventual-Inc/Daft/issues/3925



from __future__ import annotations

import os
import pathlib
import shutil
import uuid

import pytest
import s3fs

import daft


@pytest.fixture(scope="function")
def bucket(minio_io_config):
    BUCKET = "write-modes-bucket"

    fs = s3fs.S3FileSystem(
        key=minio_io_config.s3.key_id,
        password=minio_io_config.s3.access_key,
        client_kwargs={"endpoint_url": minio_io_config.s3.endpoint_url},
    )
    if not fs.exists(BUCKET):
        fs.mkdir(BUCKET)
    yield BUCKET


@pytest.fixture(scope="function")
def tmp_relative_path():
    # Create a unique directory in the home directory
    base_relative_dir = pathlib.Path(f"~/test_dir_{uuid.uuid4()}")
    full_relative_dir = base_relative_dir.joinpath("relative", "path")

    # Return the ~ version for the test to use
    yield full_relative_dir

    # Remove the directory
    full_path = os.path.expanduser(base_relative_dir)
    shutil.rmtree(full_path)


def write(
    df: daft.DataFrame,
    path: str,
    format: str,
    write_mode: str,
    partition_cols: list[str] | None = None,
    io_config: daft.io.IOConfig | None = None,
):
    if format == "parquet":
        return df.write_parquet(
            path,
            write_mode=write_mode,
            partition_cols=partition_cols,
            io_config=io_config,
        )
    elif format == "csv":
        return df.write_csv(
            path,
            write_mode=write_mode,
            partition_cols=partition_cols,
            io_config=io_config,
        )
    else:
        raise ValueError(f"Unsupported format: {format}")


def read(path: str, format: str, io_config: daft.io.IOConfig | None = None):
    if format == "parquet":
        return daft.read_parquet(path, io_config=io_config)
    elif format == "csv":
        return daft.read_csv(path, io_config=io_config)
    else:
        raise ValueError(f"Unsupported format: {format}")


def arrange_write_mode_test(existing_data, new_data, path, format, write_mode, partition_cols, io_config):
    # Write some existing_data
    write(existing_data, path, format, "append", partition_cols, io_config)

    # Write some new data
    write(new_data, path, format, write_mode, partition_cols, io_config)

    # Read back the data
    read_path = path + "/**" if partition_cols is not None else path
    read_back = read(read_path, format, io_config).sort(["a", "b"]).to_pydict()

    return read_back


def _run_append_overwrite_test(
    path,
    write_mode,
    format,
    num_partitions,
    partition_cols,
    io_config,
):
    existing_data = {"a": ["a", "a", "b", "b"], "b": [1, 2, 3, 4]}
    new_data = {
        "a": ["a", "a", "b", "b"],
        "b": [5, 6, 7, 8],
    }

    read_back = arrange_write_mode_test(
        daft.from_pydict(existing_data).into_partitions(num_partitions),
        daft.from_pydict(new_data).into_partitions(num_partitions),
        path,
        format,
        write_mode,
        partition_cols,
        io_config,
    )

    # Check the data
    if write_mode == "append":
        assert read_back["a"] == ["a"] * 4 + ["b"] * 4
        assert read_back["b"] == [1, 2, 5, 6, 3, 4, 7, 8]
    elif write_mode == "overwrite":
        assert read_back["a"] == ["a", "a", "b", "b"]
        assert read_back["b"] == [5, 6, 7, 8]
    else:
        raise ValueError(f"Unsupported write_mode: {write_mode}")


@pytest.mark.parametrize("write_mode", ["append", "overwrite"])
@pytest.mark.parametrize("format", ["csv", "parquet"])
@pytest.mark.parametrize("num_partitions", [1, 2])
@pytest.mark.parametrize("partition_cols", [None, ["a"]])
def test_append_and_overwrite_local(tmp_path, write_mode, format, num_partitions, partition_cols):
    _run_append_overwrite_test(
        path=str(tmp_path),
        write_mode=write_mode,
        format=format,
        num_partitions=num_partitions,
        partition_cols=partition_cols,
        io_config=None,
    )


@pytest.mark.parametrize("write_mode", ["append", "overwrite"])
@pytest.mark.parametrize("format", ["csv", "parquet"])
@pytest.mark.parametrize("num_partitions", [1, 2])
@pytest.mark.parametrize("partition_cols", [None, ["a"]])
def test_append_and_overwrite_local_relative_path(
    tmp_relative_path, write_mode, format, num_partitions, partition_cols
):
    _run_append_overwrite_test(
        path=str(tmp_relative_path),
        write_mode=write_mode,
        format=format,
        num_partitions=num_partitions,
        partition_cols=partition_cols,
        io_config=None,
    )


@pytest.mark.integration()
@pytest.mark.parametrize("write_mode", ["append", "overwrite"])
@pytest.mark.parametrize("format", ["csv", "parquet"])
@pytest.mark.parametrize("num_partitions", [1, 2])
@pytest.mark.parametrize("partition_cols", [None, ["a"]])
def test_append_and_overwrite_s3_minio(
    minio_io_config,
    bucket,
    write_mode,
    format,
    num_partitions,
    partition_cols,
):
    _run_append_overwrite_test(
        path=f"s3://{bucket}/{uuid.uuid4()!s}",
        write_mode=write_mode,
        format=format,
        num_partitions=num_partitions,
        partition_cols=partition_cols,
        io_config=minio_io_config,
    )


def _run_write_modes_empty_test(
    path,
    write_mode,
    format,
    io_config,
):
    existing_data = {"a": ["a", "a", "b", "b"], "b": ["c", "d", "e", "f"]}
    new_data = {
        "a": ["a", "a", "b", "b"],
        "b": ["g", "h", "i", "j"],
    }

    read_back = arrange_write_mode_test(
        daft.from_pydict(existing_data),
        daft.from_pydict(new_data).where(daft.lit(False)),  # Empty data
        path,
        format,
        write_mode,
        None,
        io_config,
    )

    # Check the data
    if write_mode == "append":
        # The data should be the same as the existing data
        assert read_back["a"] == ["a", "a", "b", "b"]
        assert read_back["b"] == ["c", "d", "e", "f"]
    elif write_mode == "overwrite":
        # The data should be empty because we are overwriting the existing data
        assert read_back["a"] == []
        assert read_back["b"] == []
    else:
        raise ValueError(f"Unsupported write_mode: {write_mode}")


@pytest.mark.parametrize("write_mode", ["append", "overwrite"])
@pytest.mark.parametrize("format", ["csv", "parquet"])
def test_write_modes_local_empty_data(tmp_path, write_mode, format):
    _run_write_modes_empty_test(
        path=str(tmp_path),
        write_mode=write_mode,
        format=format,
        io_config=None,
    )


@pytest.mark.integration()
@pytest.mark.parametrize("write_mode", ["append", "overwrite"])
@pytest.mark.parametrize("format", ["csv", "parquet"])
def test_write_modes_s3_minio_empty_data(
    minio_io_config,
    bucket,
    write_mode,
    format,
):
    _run_write_modes_empty_test(
        path=f"s3://{bucket}/{uuid.uuid4()!s}",
        write_mode=write_mode,
        format=format,
        io_config=minio_io_config,
    )


OVERWRITE_PARTITION_TEST_CASES = [
    pytest.param(
        {
            "a": ["a", "a", "b", "b"],
            "b": [5, 6, 7, 8],
        },
        {
            "a": ["a", "a", "b", "b"],
            "b": [5, 6, 7, 8],
        },
        id="overwrite-all",
    ),
    pytest.param(
        {
            "a": ["a", "a"],
            "b": [5, 6],
        },
        {
            "a": ["a", "a", "b", "b"],
            "b": [5, 6, 3, 4],
        },
        id="overwrite-some",
    ),
    pytest.param(
        {
            "a": ["b", "b", "c", "c"],
            "b": [9, 10, 11, 12],
        },
        {
            "a": ["a", "a", "b", "b", "c", "c"],
            "b": [1, 2, 9, 10, 11, 12],
        },
        id="overwrite-and-append",
    ),
]


def _run_overwrite_partitions_test(
    path,
    format,
    new_data,
    expected_read_back,
    io_config,
):
    existing_data = {"a": ["a", "a", "b", "b"], "b": [1, 2, 3, 4]}

    read_back = arrange_write_mode_test(
        daft.from_pydict(existing_data),
        daft.from_pydict(new_data),
        path,
        format,
        "overwrite-partitions",
        ["a"],
        io_config,
    )

    # Check the data
    for col in expected_read_back:
        assert read_back[col] == expected_read_back[col]


@pytest.mark.parametrize("format", ["csv", "parquet"])
@pytest.mark.parametrize("new_data, expected_read_back", OVERWRITE_PARTITION_TEST_CASES)
def test_overwrite_partitions_local(tmp_path, format, new_data, expected_read_back):
    _run_overwrite_partitions_test(
        path=str(tmp_path),
        format=format,
        new_data=new_data,
        expected_read_back=expected_read_back,
        io_config=None,
    )


@pytest.mark.parametrize("format", ["csv", "parquet"])
@pytest.mark.parametrize("new_data, expected_read_back", OVERWRITE_PARTITION_TEST_CASES)
def test_overwrite_partitions_local_relative_path(tmp_relative_path, format, new_data, expected_read_back):
    _run_overwrite_partitions_test(
        path=str(tmp_relative_path),
        format=format,
        new_data=new_data,
        expected_read_back=expected_read_back,
        io_config=None,
    )


@pytest.mark.integration()
@pytest.mark.parametrize("format", ["csv", "parquet"])
@pytest.mark.parametrize("new_data, expected_read_back", OVERWRITE_PARTITION_TEST_CASES)
def test_overwrite_partitions_s3_minio(
    minio_io_config,
    bucket,
    format,
    new_data,
    expected_read_back,
):
    _run_overwrite_partitions_test(
        path=f"s3://{bucket}/{uuid.uuid4()!s}",
        format=format,
        new_data=new_data,
        expected_read_back=expected_read_back,
        io_config=minio_io_config,
    )
